# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import pathlib
import tempfile
import textwrap
from copy import deepcopy
from glob import glob
from typing import TYPE_CHECKING

import numpy as np
import torch

from monai.apps import download_and_extract
from monai.transforms import (
    Affine,
    Affined,
    AsDiscrete,
    Compose,
    EnsureChannelFirstd,
    Flip,
    Flipd,
    LoadImaged,
    MapTransform,
    Orientation,
    Orientationd,
    Rand3DElastic,
    Rand3DElasticd,
    RandFlip,
    RandFlipd,
    Randomizable,
    RandRotate,
    RandRotated,
    RandZoom,
    RandZoomd,
    Rotate,
    Rotate90,
    Rotate90d,
    Rotated,
    ScaleIntensity,
    ScaleIntensityd,
    SpatialPadd,
    Zoom,
    Zoomd,
)
from monai.transforms.croppad.array import (
    BorderPad,
    CenterScaleCrop,
    CenterSpatialCrop,
    CropForeground,
    DivisiblePad,
    RandCropByLabelClasses,
    RandCropByPosNegLabel,
    RandScaleCrop,
    RandSpatialCrop,
    RandSpatialCropSamples,
    RandWeightedCrop,
    ResizeWithPadOrCrop,
    SpatialCrop,
    SpatialPad,
)
from monai.transforms.croppad.dictionary import (
    BorderPadd,
    CenterScaleCropd,
    CenterSpatialCropd,
    CropForegroundd,
    DivisiblePadd,
    RandCropByLabelClassesd,
    RandCropByPosNegLabeld,
    RandScaleCropd,
    RandSpatialCropd,
    RandSpatialCropSamplesd,
    RandWeightedCropd,
    ResizeWithPadOrCropd,
    SpatialCropd,
)
from monai.transforms.intensity.array import (
    AdjustContrast,
    ForegroundMask,
    GaussianSharpen,
    GaussianSmooth,
    GibbsNoise,
    HistogramNormalize,
    KSpaceSpikeNoise,
    MaskIntensity,
    MedianSmooth,
    NormalizeIntensity,
    RandAdjustContrast,
    RandBiasField,
    RandCoarseDropout,
    RandCoarseShuffle,
    RandGaussianNoise,
    RandGaussianSharpen,
    RandGaussianSmooth,
    RandGibbsNoise,
    RandHistogramShift,
    RandKSpaceSpikeNoise,
    RandRicianNoise,
    RandScaleIntensity,
    RandShiftIntensity,
    RandStdShiftIntensity,
    SavitzkyGolaySmooth,
    ScaleIntensityRange,
    ScaleIntensityRangePercentiles,
    ShiftIntensity,
    StdShiftIntensity,
    ThresholdIntensity,
)
from monai.transforms.intensity.dictionary import (
    AdjustContrastd,
    ForegroundMaskd,
    GaussianSharpend,
    GaussianSmoothd,
    GibbsNoised,
    HistogramNormalized,
    KSpaceSpikeNoised,
    MaskIntensityd,
    MedianSmoothD,
    NormalizeIntensityd,
    RandAdjustContrastd,
    RandBiasFieldd,
    RandCoarseDropoutd,
    RandCoarseShuffled,
    RandGaussianNoised,
    RandGaussianSharpend,
    RandGaussianSmoothd,
    RandGibbsNoised,
    RandHistogramShiftd,
    RandKSpaceSpikeNoised,
    RandRicianNoised,
    RandScaleIntensityd,
    RandShiftIntensityd,
    RandStdShiftIntensityd,
    SavitzkyGolaySmoothd,
    ScaleIntensityRanged,
    ScaleIntensityRangePercentilesd,
    ShiftIntensityd,
    StdShiftIntensityd,
    ThresholdIntensityd,
)
from monai.transforms.post.array import KeepLargestConnectedComponent, LabelFilter, LabelToContour, RemoveSmallObjects
from monai.transforms.post.dictionary import (
    AsDiscreted,
    KeepLargestConnectedComponentd,
    LabelFilterd,
    LabelToContourd,
    RemoveSmallObjectsd,
)
from monai.transforms.smooth_field.array import (
    RandSmoothDeform,
    RandSmoothFieldAdjustContrast,
    RandSmoothFieldAdjustIntensity,
)
from monai.transforms.smooth_field.dictionary import (
    RandSmoothDeformd,
    RandSmoothFieldAdjustContrastd,
    RandSmoothFieldAdjustIntensityd,
)
from monai.transforms.spatial.array import (
    GridDistortion,
    Rand2DElastic,
    RandAffine,
    RandAxisFlip,
    RandGridDistortion,
    RandRotate90,
    Resize,
    Spacing,
)
from monai.transforms.spatial.dictionary import (
    GridDistortiond,
    Rand2DElasticd,
    RandAffined,
    RandAxisFlipd,
    RandGridDistortiond,
    RandRotate90d,
    Resized,
    Spacingd,
)
from monai.utils.enums import CommonKeys
from monai.utils.misc import MONAIEnvVars
from monai.utils.module import optional_import

if TYPE_CHECKING:
    import matplotlib.pyplot as plt

    has_matplotlib = True

else:
    plt, has_matplotlib = optional_import("matplotlib.pyplot")


def get_data(keys):
    """Get the example data to be used.

    Use MarsAtlas as it only contains 1 image for quick download and
    that image is parcellated.
    """
    cache_dir = MONAIEnvVars.data_dir() or tempfile.mkdtemp()
    fname = "MarsAtlas-MNI-Colin27.zip"
    url = "https://www.dropbox.com/s/ndz8qtqblkciole/" + fname + "?dl=1"
    out_path = os.path.join(cache_dir, "MarsAtlas-MNI-Colin27")
    zip_path = os.path.join(cache_dir, fname)

    download_and_extract(url, zip_path, out_path)

    image, label = sorted(glob(os.path.join(out_path, "*.nii")))

    transforms = Compose(
        [
            LoadImaged(keys),
            EnsureChannelFirstd(keys),
            ScaleIntensityd(CommonKeys.IMAGE),
            Rotate90d(keys, spatial_axes=(0, 2)),
        ]
    )
    data = transforms({CommonKeys.IMAGE: image, CommonKeys.LABEL: label})
    max_size = max(data[keys[0]].shape)
    padder = SpatialPadd(keys, (max_size, max_size, max_size))
    return padder(data)


def update_docstring(code_path, transform_name):
    """
    Find the documentation for a given transform and if it's missing,
    add a pointer to the transform's example image.
    """
    with open(code_path) as f:
        contents = f.readlines()
    doc_start = None
    for i, line in enumerate(contents):
        # find the line containing start of the transform documentation
        if "`" + transform_name + "`" in line:
            doc_start = i
            break
    if doc_start is None:
        raise RuntimeError("Couldn't find transform documentation")

    # if image is already in docs, nothing to do
    image_line = doc_start + 2
    if ".. image" in contents[image_line]:
        return

    # add the line for the image and the alt text
    contents_orig = deepcopy(contents)
    contents.insert(
        image_line,
        ".. image:: https://github.com/Project-MONAI/DocImages/raw/main/transforms/" + transform_name + ".png\n",
    )
    contents.insert(image_line + 1, "    :alt: example of " + transform_name + "\n")

    # check that we've only added two lines
    if len(contents) != len(contents_orig) + 2:
        raise AssertionError

    # write the updated doc to overwrite the original
    with open(code_path, "w") as f:
        f.writelines(contents)


def pre_process_data(data, ndim, is_map, is_post):
    """If transform requires 2D data, then convert to 2D by selecting the middle of the last dimension."""
    if ndim == 2:
        data = {k: v[..., v.shape[-1] // 2] for k, v in data.items()}
    if is_map:
        return data
    return data[CommonKeys.LABEL] if is_post else data[CommonKeys.IMAGE]


def get_2d_slice(image, view: int, is_label):
    """If image is 3d, get the central slice. If is already 2d, return as-is.
    If image is label, set 0 to np.nan.
    """
    if image.ndim == 2:
        out = image
    else:
        shape = image.shape
        slices = [slice(0, s) for s in shape]
        _slice = shape[view] // 2
        slices[view] = slice(_slice, _slice + 1)
        out = np.squeeze(image[tuple(slices)], view)
    if is_label:
        out[out == 0] = np.nan
    return out


def get_stacked_2d_ims(im, is_label):
    """Get the 3 orthogonal views and stack them into 1 image.
    Requires that all images be same size, but this is taken care
    of by the `SpatialPadd` earlier.
    """
    return [get_2d_slice(im, i, is_label) for i in range(3)]


def get_stacked_before_after(before, after, is_label=False):
    """Stack before and after images into 1 image if 3d.
    Requires that before and after images be the same size.
    """
    return [get_stacked_2d_ims(d, is_label) for d in (before, after)]


def save_image(images, labels, filename, transform_name, transform_args, shapes, colorbar=False):
    """Save image to file, ensuring there's no whitespace around the edge."""
    plt.rcParams.update({"font.family": "monospace"})
    plt.style.use("dark_background")
    nrow = len(images)  # before and after (should always be 2)
    ncol = len(images[0])  # num orthogonal views (either 1 or 3)
    # roughly estimate the height_ratios of the first:second row
    hs = [float(r[0].shape[0]) for r in images]
    fig = plt.figure(tight_layout=True)
    spec = fig.add_gridspec(nrow, ncol, hspace=0, wspace=0, height_ratios=hs)
    for row in range(nrow):
        vmin = min(i.min() for i in images[row])
        vmax = max(i.max() for i in images[row])
        for col in range(ncol):
            ax = fig.add_subplot(spec[row, col])
            imshow = ax.imshow(images[row][col], cmap="gray", vmin=vmin, vmax=vmax)
            ax.set_aspect("equal")
            if colorbar and col == ncol - 1:
                plt.colorbar(imshow, ax=ax)
            if col == 0:
                y_label = "After" if row else "Before"
                y_label += ("\n" + shapes[row]) if shapes[0] != shapes[1] else ""
                ax.set_ylabel(y_label)
            # print yticks for the right most column
            if col != ncol - 1 or colorbar:
                ax.set_yticks([])
            else:
                ax.yaxis.tick_right()
                for n, label in enumerate(ax.yaxis.get_ticklabels()):
                    if n > 2:
                        label.set_visible(False)
            ax.set_xticks([])
            ax.set_frame_on(False)
            if labels is not None:
                ax.imshow(labels[row][col], cmap="hsv", alpha=0.9, interpolation="nearest")
    # title is e.g., Flipd(keys=keys, spatial_axis=0)
    title = transform_name + "("
    for k, v in transform_args.items():
        title += k + "="
        if isinstance(v, str):
            title += "'" + v + "'"
        elif isinstance(v, (np.ndarray, torch.Tensor)):
            title += "[array]"
        elif callable(v):
            title += "[callable]"
        else:
            title += str(v)
        title += ", "
    if len(transform_args) > 0:
        title = title[:-2]
    title += ")"
    # shorten the lines
    title = textwrap.fill(title, 50, break_long_words=False, subsequent_indent=" " * (len(transform_name) + 1))
    fig.suptitle(title, x=0.1, horizontalalignment="left")
    fig.savefig(filename)
    plt.close(fig)


def get_images(data, is_label=False):
    """Get image. If is dictionary, extract key. If is list, stack. If both dictionary and list, do both.
    Also return the image size as string to be used im the imshow. If it's a list, return `N x (H,W,D)`.
    """
    # If not a list, convert
    if not isinstance(data, list):
        data = [data]
    key = CommonKeys.LABEL if is_label else CommonKeys.IMAGE
    is_map = isinstance(data[0], dict)
    # length of the list will be equal to number of samples produced. This will be 1 except for transforms that
    # produce `num_samples`.
    data = [d[key] if is_map else d for d in data]
    data = [d[0] for d in data]  # remove channel component

    # for each sample, create a list of the orthogonal views. If image is 2d, length will be 1. If 3d, there
    # will be three orthogonal views
    num_samples = len(data)
    num_orthog_views = 3 if data[0].ndim == 3 else 1
    shape_str = (f"{num_samples} x " if num_samples > 1 else "") + str(data[0].shape)
    for i in range(num_samples):
        data[i] = [get_2d_slice(data[i], view, is_label) for view in range(num_orthog_views)]

    out = []
    if num_samples == 1:
        out = data[0]
    else:
        # we might need to panel the images. this happens if a transform produces e.g. 4 output images.
        # In this case, we create a 2-by-2 grid from them. Output will be a list containing n_orthog_views,
        # each element being either the image (if num_samples is 1) or the panelled image.
        nrows = int(np.floor(num_samples**0.5))
        for view in range(num_orthog_views):
            result = np.asarray([d[view] for d in data])
            nindex, height, width = result.shape
            ncols = nindex // nrows
            # only implemented for square number of images (e.g. 4 images goes to a 2-by-2 panel)
            if nindex != nrows * ncols:
                raise NotImplementedError
            # want result.shape = (height*nrows, width*ncols), have to be careful about striding
            result = result.reshape(nrows, ncols, height, width).swapaxes(1, 2).reshape(height * nrows, width * ncols)
            out.append(result)
    return out, shape_str


def create_transform_im(
    transform, transform_args, data, ndim=3, colorbar=False, update_doc=True, seed=0, is_post=False
):
    """Create an image with the before and after of the transform.
    Also update the transform's documentation to point to this image."""

    transform = transform(**transform_args)

    if not has_matplotlib:
        raise RuntimeError

    if isinstance(transform, Randomizable):
        # increment the seed for map transforms so they're different to the array versions.
        seed = seed + 1 if isinstance(transform, MapTransform) else seed
        transform.set_random_state(seed)

    out_dir = MONAIEnvVars.doc_images()
    if out_dir is None:
        raise RuntimeError(
            "Please git clone https://github.com/Project-MONAI/DocImages"
            + " and then set the environment variable `MONAI_DOC_IMAGES`"
        )
    out_dir = os.path.join(out_dir, "transforms")

    # Path is transform name
    transform_name = transform.__class__.__name__
    out_fname = transform_name + ".png"
    out_file = os.path.join(out_dir, out_fname)

    is_map = isinstance(transform, MapTransform)
    data_in = pre_process_data(deepcopy(data), ndim, is_map, is_post)

    data_tr = transform(deepcopy(data_in))

    images_before, before_shape = get_images(data_in)
    images_after, after_shape = get_images(data_tr)
    images = (images_before, images_after)
    shapes = (before_shape, after_shape)

    labels = None
    if is_map:
        labels_before, *_ = get_images(data_in, is_label=True)
        labels_after, *_ = get_images(data_tr, is_label=True)
        labels = (labels_before, labels_after)

    save_image(images, labels, out_file, transform_name, transform_args, shapes, colorbar)

    if update_doc:
        base_dir = pathlib.Path(__file__).parent.parent.parent
        rst_path = os.path.join(base_dir, "docs", "source", "transforms.rst")
        update_docstring(rst_path, transform_name)


if __name__ == "__main__":
    keys = [CommonKeys.IMAGE, CommonKeys.LABEL]
    data = get_data(keys)
    create_transform_im(RandFlip, dict(prob=1, spatial_axis=1), data)
    create_transform_im(RandFlipd, dict(keys=keys, prob=1, spatial_axis=2), data)
    create_transform_im(Flip, dict(spatial_axis=1), data)
    create_transform_im(Flipd, dict(keys=keys, spatial_axis=2), data)
    create_transform_im(Orientation, dict(axcodes="RPI"), data)
    create_transform_im(Orientationd, dict(keys=keys, axcodes="RPI"), data)
    create_transform_im(
        Rand3DElastic, dict(prob=1.0, sigma_range=(1, 2), magnitude_range=(0.5, 0.5), shear_range=(1, 1, 1)), data
    )
    create_transform_im(Affine, dict(shear_params=(0, 0.5, 0), image_only=True, padding_mode="zeros"), data)
    create_transform_im(
        Affined, dict(keys=keys, shear_params=(0, 0.5, 0), mode=["bilinear", "nearest"], padding_mode="zeros"), data
    )
    create_transform_im(RandAffine, dict(prob=1, shear_range=(0.5, 0.5), padding_mode="zeros"), data)
    create_transform_im(
        RandAffined,
        dict(keys=keys, prob=1, shear_range=(0.5, 0.5), mode=["bilinear", "nearest"], padding_mode="zeros"),
        data,
    )
    create_transform_im(
        Rand3DElastic, dict(sigma_range=(5, 7), magnitude_range=(50, 150), prob=1, padding_mode="zeros"), data
    )
    create_transform_im(
        Rand2DElastic, dict(prob=1, spacing=(20, 20), magnitude_range=(1, 2), padding_mode="zeros"), data, 2
    )
    create_transform_im(
        Rand2DElasticd,
        dict(
            keys=keys,
            prob=1,
            spacing=(20, 20),
            magnitude_range=(1, 2),
            padding_mode="zeros",
            mode=["bilinear", "nearest"],
        ),
        data,
        2,
    )
    create_transform_im(
        Rand3DElasticd,
        dict(
            keys=keys,
            sigma_range=(5, 7),
            magnitude_range=(50, 150),
            prob=1,
            padding_mode="zeros",
            mode=["bilinear", "nearest"],
        ),
        data,
    )
    create_transform_im(Rotate90, dict(spatial_axes=(1, 2)), data)
    create_transform_im(Rotate90d, dict(keys=keys, spatial_axes=(1, 2)), data)
    create_transform_im(RandRotate90, dict(prob=1), data)
    create_transform_im(RandRotate90d, dict(keys=keys, prob=1), data)
    create_transform_im(Rotate, dict(angle=0.1), data)
    create_transform_im(Rotated, dict(keys=keys, angle=0.1, mode=["bilinear", "nearest"]), data)
    create_transform_im(RandRotate, dict(prob=1, range_x=[0.4, 0.4]), data)
    create_transform_im(RandRotated, dict(keys=keys, prob=1, range_x=[0.4, 0.4], mode=["bilinear", "nearest"]), data)
    create_transform_im(Zoom, dict(zoom=0.6), data)
    create_transform_im(Zoomd, dict(keys=keys, zoom=1.3, mode=["area", "nearest"]), data)
    create_transform_im(RandZoom, dict(prob=1, min_zoom=0.6, max_zoom=0.8), data)
    create_transform_im(RandZoomd, dict(keys=keys, prob=1, min_zoom=1.3, max_zoom=1.5, mode=["area", "nearest"]), data)
    create_transform_im(ScaleIntensity, dict(minv=0, maxv=10), data, colorbar=True)
    create_transform_im(ScaleIntensityd, dict(keys=CommonKeys.IMAGE, minv=0, maxv=10), data, colorbar=True)
    create_transform_im(RandScaleIntensity, dict(prob=1.0, factors=(5, 10)), data, colorbar=True)
    create_transform_im(
        RandScaleIntensityd, dict(keys=CommonKeys.IMAGE, prob=1.0, factors=(5, 10)), data, colorbar=True
    )
    create_transform_im(DivisiblePad, dict(k=64), data)
    create_transform_im(DivisiblePadd, dict(keys=keys, k=64), data)
    create_transform_im(CropForeground, dict(allow_smaller=False), data)
    create_transform_im(CropForegroundd, dict(keys=keys, source_key=CommonKeys.IMAGE, allow_smaller=False), data)
    create_transform_im(RandGaussianNoise, dict(prob=1, mean=0, std=0.1), data)
    create_transform_im(RandGaussianNoised, dict(keys=CommonKeys.IMAGE, prob=1, mean=0, std=0.1), data)
    create_transform_im(KSpaceSpikeNoise, dict(loc=(100, 100, 100), k_intensity=13), data)
    create_transform_im(KSpaceSpikeNoised, dict(keys=CommonKeys.IMAGE, loc=(100, 100, 100), k_intensity=13), data)
    create_transform_im(RandKSpaceSpikeNoise, dict(prob=1, intensity_range=(10, 13)), data)
    create_transform_im(RandKSpaceSpikeNoised, dict(keys=CommonKeys.IMAGE, prob=1, intensity_range=(13, 15)), data)
    create_transform_im(RandRicianNoise, dict(prob=1.0, mean=1, std=0.5), data)
    create_transform_im(RandRicianNoised, dict(keys=CommonKeys.IMAGE, prob=1.0, mean=1, std=0.5), data)
    create_transform_im(SavitzkyGolaySmooth, dict(window_length=5, order=1), data)
    create_transform_im(SavitzkyGolaySmoothd, dict(keys=CommonKeys.IMAGE, window_length=5, order=1), data)
    create_transform_im(GibbsNoise, dict(alpha=0.8), data)
    create_transform_im(GibbsNoised, dict(keys=CommonKeys.IMAGE, alpha=0.8), data)
    create_transform_im(RandGibbsNoise, dict(prob=1.0, alpha=(0.6, 0.8)), data)
    create_transform_im(RandGibbsNoised, dict(keys=CommonKeys.IMAGE, prob=1.0, alpha=(0.6, 0.8)), data)
    create_transform_im(ShiftIntensity, dict(offset=1), data, colorbar=True)
    create_transform_im(ShiftIntensityd, dict(keys=CommonKeys.IMAGE, offset=1), data, colorbar=True)
    create_transform_im(RandShiftIntensity, dict(prob=1.0, offsets=(10, 20)), data, colorbar=True)
    create_transform_im(
        RandShiftIntensityd, dict(keys=CommonKeys.IMAGE, prob=1.0, offsets=(10, 20)), data, colorbar=True
    )
    create_transform_im(StdShiftIntensity, dict(factor=10), data, colorbar=True)
    create_transform_im(StdShiftIntensityd, dict(keys=CommonKeys.IMAGE, factor=10), data, colorbar=True)
    create_transform_im(RandStdShiftIntensity, dict(prob=1.0, factors=(5, 10)), data, colorbar=True)
    create_transform_im(
        RandStdShiftIntensityd, dict(keys=CommonKeys.IMAGE, prob=1.0, factors=(5, 10)), data, colorbar=True
    )
    create_transform_im(RandBiasField, dict(prob=1, coeff_range=(0.2, 0.3)), data)
    create_transform_im(RandBiasFieldd, dict(keys=CommonKeys.IMAGE, prob=1, coeff_range=(0.2, 0.3)), data)
    create_transform_im(NormalizeIntensity, dict(subtrahend=0, divisor=10), data, colorbar=True)
    create_transform_im(NormalizeIntensityd, dict(keys=CommonKeys.IMAGE, subtrahend=0, divisor=10), data, colorbar=True)
    create_transform_im(ThresholdIntensity, dict(threshold=0.4, above=False, cval=0.9), data, colorbar=True)
    create_transform_im(
        ThresholdIntensityd, dict(keys=CommonKeys.IMAGE, threshold=0.4, above=False, cval=0.9), data, colorbar=True
    )
    create_transform_im(ScaleIntensityRange, dict(a_min=0, a_max=1, b_min=1, b_max=10), data, colorbar=True)
    create_transform_im(
        ScaleIntensityRanged, dict(keys=CommonKeys.IMAGE, a_min=0, a_max=1, b_min=1, b_max=10), data, colorbar=True
    )
    create_transform_im(ScaleIntensityRangePercentiles, dict(lower=5, upper=95, b_min=1, b_max=10), data, colorbar=True)
    create_transform_im(
        ScaleIntensityRangePercentilesd,
        dict(keys=CommonKeys.IMAGE, lower=5, upper=95, b_min=1, b_max=10),
        data,
        colorbar=True,
    )
    create_transform_im(AdjustContrast, dict(gamma=2), data, colorbar=True)
    create_transform_im(AdjustContrastd, dict(keys=CommonKeys.IMAGE, gamma=2), data, colorbar=True)
    create_transform_im(RandAdjustContrast, dict(prob=1, gamma=(1.5, 2)), data, colorbar=True)
    create_transform_im(RandAdjustContrastd, dict(keys=CommonKeys.IMAGE, prob=1, gamma=(1.5, 2)), data, colorbar=True)
    create_transform_im(MaskIntensity, dict(mask_data=data[CommonKeys.IMAGE], select_fn=lambda x: x > 0.3), data)
    create_transform_im(
        MaskIntensityd, dict(keys=CommonKeys.IMAGE, mask_key=CommonKeys.IMAGE, select_fn=lambda x: x > 0.3), data
    )
    create_transform_im(ForegroundMask, dict(invert=True), data)
    create_transform_im(ForegroundMaskd, dict(keys=CommonKeys.IMAGE, invert=True), data)
    create_transform_im(GaussianSmooth, dict(sigma=2), data)
    create_transform_im(GaussianSmoothd, dict(keys=CommonKeys.IMAGE, sigma=2), data)
    create_transform_im(MedianSmooth, dict(radius=3), data)
    create_transform_im(MedianSmoothD, dict(keys=keys, radius=1), data)
    create_transform_im(RandGaussianSmooth, dict(prob=1.0, sigma_x=(1, 2)), data)
    create_transform_im(RandGaussianSmoothd, dict(keys=CommonKeys.IMAGE, prob=1.0, sigma_x=(1, 2)), data)
    create_transform_im(GaussianSharpen, dict(), GaussianSmoothd(CommonKeys.IMAGE, 2)(data))
    create_transform_im(GaussianSharpend, dict(keys=CommonKeys.IMAGE), GaussianSmoothd(CommonKeys.IMAGE, 2)(data))
    create_transform_im(RandGaussianSharpen, dict(prob=1), GaussianSmoothd(CommonKeys.IMAGE, 2)(data))
    create_transform_im(
        RandGaussianSharpend, dict(keys=CommonKeys.IMAGE, prob=1), GaussianSmoothd(CommonKeys.IMAGE, 2)(data)
    )
    create_transform_im(RandHistogramShift, dict(prob=1, num_control_points=3), data, colorbar=True)
    create_transform_im(
        RandHistogramShiftd, dict(keys=CommonKeys.IMAGE, prob=1, num_control_points=3), data, colorbar=True
    )
    create_transform_im(RandCoarseDropout, dict(prob=1, holes=200, spatial_size=20, fill_value=0), data)
    create_transform_im(
        RandCoarseDropoutd, dict(keys=CommonKeys.IMAGE, prob=1, holes=200, spatial_size=20, fill_value=0), data
    )
    create_transform_im(RandCoarseShuffle, dict(prob=1, holes=200, spatial_size=20), data)
    create_transform_im(RandCoarseShuffled, dict(keys=CommonKeys.IMAGE, prob=1, holes=200, spatial_size=20), data)
    create_transform_im(HistogramNormalize, dict(num_bins=10), data)
    create_transform_im(HistogramNormalized, dict(keys=CommonKeys.IMAGE, num_bins=10), data)
    create_transform_im(SpatialPad, dict(spatial_size=(300, 300, 300)), data)
    create_transform_im(SpatialPadd, dict(keys=keys, spatial_size=(300, 300, 300)), data)
    create_transform_im(BorderPad, dict(spatial_border=10), data)
    create_transform_im(BorderPadd, dict(keys=keys, spatial_border=10), data)
    create_transform_im(SpatialCrop, dict(roi_center=(75, 75, 75), roi_size=(100, 100, 100)), data)
    create_transform_im(SpatialCropd, dict(keys=keys, roi_center=(75, 75, 75), roi_size=(100, 100, 100)), data)
    create_transform_im(CenterSpatialCrop, dict(roi_size=(100, 100, 100)), data)
    create_transform_im(CenterSpatialCropd, dict(keys=keys, roi_size=(100, 100, 100)), data)
    create_transform_im(RandSpatialCrop, dict(roi_size=(100, 100, 100), random_size=False), data)
    create_transform_im(RandSpatialCropd, dict(keys=keys, roi_size=(100, 100, 100), random_size=False), data)
    create_transform_im(RandSpatialCropSamples, dict(num_samples=4, roi_size=(100, 100, 100), random_size=False), data)
    create_transform_im(
        RandSpatialCropSamplesd, dict(keys=keys, num_samples=4, roi_size=(100, 100, 100), random_size=False), data
    )
    create_transform_im(
        RandWeightedCrop, dict(spatial_size=(100, 100, 100), num_samples=4, weight_map=data[CommonKeys.IMAGE] > 0), data
    )
    create_transform_im(
        RandWeightedCropd, dict(keys=keys, spatial_size=(100, 100, 100), num_samples=4, w_key=CommonKeys.IMAGE), data
    )
    create_transform_im(
        RandCropByPosNegLabel,
        dict(spatial_size=(100, 100, 100), label=data[CommonKeys.LABEL], neg=0, num_samples=4),
        data,
    )
    create_transform_im(
        RandCropByPosNegLabeld,
        dict(keys=keys, spatial_size=(100, 100, 100), label_key=CommonKeys.LABEL, neg=0, num_samples=4),
        data,
    )
    create_transform_im(
        RandCropByLabelClasses,
        dict(
            spatial_size=(100, 100, 100), label=data[CommonKeys.LABEL] > 0, num_classes=2, ratios=[0, 1], num_samples=4
        ),
        data,
    )
    create_transform_im(
        RandCropByLabelClassesd,
        dict(
            keys=keys,
            spatial_size=(100, 100, 100),
            label_key=CommonKeys.LABEL,
            num_classes=2,
            ratios=[0, 1],
            num_samples=4,
        ),
        data,
    )
    create_transform_im(ResizeWithPadOrCrop, dict(spatial_size=(100, 100, 100)), data)
    create_transform_im(ResizeWithPadOrCropd, dict(keys=keys, spatial_size=(100, 100, 100)), data)
    create_transform_im(RandScaleCrop, dict(roi_scale=0.4), data)
    create_transform_im(RandScaleCropd, dict(keys=keys, roi_scale=0.4), data)
    create_transform_im(CenterScaleCrop, dict(roi_scale=0.4), data)
    create_transform_im(CenterScaleCropd, dict(keys=keys, roi_scale=0.4), data)
    create_transform_im(AsDiscrete, dict(to_onehot=None, threshold=10), data, is_post=True, colorbar=True)
    create_transform_im(AsDiscreted, dict(keys=CommonKeys.LABEL, to_onehot=None, threshold=10), data, is_post=True)
    create_transform_im(LabelFilter, dict(applied_labels=(1, 2, 3, 4, 5, 6)), data, is_post=True)
    create_transform_im(
        LabelFilterd, dict(keys=CommonKeys.LABEL, applied_labels=(1, 2, 3, 4, 5, 6)), data, is_post=True
    )
    create_transform_im(LabelToContour, dict(), data, is_post=True)
    create_transform_im(LabelToContourd, dict(keys=CommonKeys.LABEL), data, is_post=True)
    create_transform_im(Spacing, dict(pixdim=(5, 5, 5)), data)
    create_transform_im(Spacingd, dict(keys=keys, pixdim=(5, 5, 5), mode=["bilinear", "nearest"]), data)
    create_transform_im(RandAxisFlip, dict(prob=1), data)
    create_transform_im(RandAxisFlipd, dict(keys=keys, prob=1), data)
    create_transform_im(Resize, dict(spatial_size=(100, 100, 100)), data)
    create_transform_im(Resized, dict(keys=keys, spatial_size=(100, 100, 100), mode=["area", "nearest"]), data)
    data_binary = deepcopy(data)
    data_binary[CommonKeys.LABEL] = (data_binary[CommonKeys.LABEL] > 0).astype(np.float32)
    create_transform_im(KeepLargestConnectedComponent, dict(applied_labels=1), data_binary, is_post=True, ndim=2)
    create_transform_im(
        KeepLargestConnectedComponentd, dict(keys=CommonKeys.LABEL, applied_labels=1), data_binary, is_post=True, ndim=2
    )
    create_transform_im(RemoveSmallObjects, dict(min_size=100), data_binary, is_post=True, ndim=2)
    create_transform_im(
        RemoveSmallObjectsd, dict(keys=CommonKeys.LABEL, min_size=100), data_binary, is_post=True, ndim=2
    )
    create_transform_im(
        GridDistortion, dict(num_cells=3, distort_steps=[(1.5,) * 4] * 3, mode="nearest", padding_mode="zeros"), data
    )
    create_transform_im(
        GridDistortiond,
        dict(
            keys=keys, num_cells=3, distort_steps=[(1.5,) * 4] * 3, mode=["bilinear", "nearest"], padding_mode="zeros"
        ),
        data,
    )
    create_transform_im(RandGridDistortion, dict(num_cells=3, prob=1.0, distort_limit=(-0.1, 0.1)), data)
    create_transform_im(
        RandGridDistortiond,
        dict(keys=keys, num_cells=4, prob=1.0, distort_limit=(-0.2, 0.2), mode=["bilinear", "nearest"]),
        data,
    )
    create_transform_im(
        RandSmoothFieldAdjustContrast, dict(spatial_size=(217, 217, 217), rand_size=(10, 10, 10), prob=1.0), data
    )
    create_transform_im(
        RandSmoothFieldAdjustContrastd,
        dict(keys=keys, spatial_size=(217, 217, 217), rand_size=(10, 10, 10), prob=1.0),
        data,
    )
    create_transform_im(
        RandSmoothFieldAdjustIntensity,
        dict(spatial_size=(217, 217, 217), rand_size=(10, 10, 10), prob=1.0, gamma=(0.5, 4.5)),
        data,
    )
    create_transform_im(
        RandSmoothFieldAdjustIntensityd,
        dict(keys=keys, spatial_size=(217, 217, 217), rand_size=(10, 10, 10), prob=1.0, gamma=(0.5, 4.5)),
        data,
    )

    create_transform_im(
        RandSmoothDeform,
        dict(spatial_size=(217, 217, 217), rand_size=(10, 10, 10), prob=1.0, def_range=0.05, grid_mode="bilinear"),
        data,
    )
    create_transform_im(
        RandSmoothDeformd,
        dict(
            keys=keys,
            spatial_size=(217, 217, 217),
            rand_size=(10, 10, 10),
            prob=1.0,
            def_range=0.05,
            grid_mode="bilinear",
        ),
        data,
    )



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections.abc import Sequence
from typing import TypeVar

import numpy as np
import torch

from monai.config.type_definitions import NdarrayOrTensor, NdarrayTensor
from monai.utils.type_conversion import convert_data_type, convert_to_dst_type

__all__ = [
    "allclose",
    "moveaxis",
    "in1d",
    "clip",
    "percentile",
    "where",
    "argwhere",
    "argsort",
    "nonzero",
    "floor_divide",
    "unravel_index",
    "unravel_indices",
    "ravel",
    "any_np_pt",
    "maximum",
    "concatenate",
    "cumsum",
    "isfinite",
    "searchsorted",
    "repeat",
    "isnan",
    "ascontiguousarray",
    "stack",
    "mode",
    "unique",
    "max",
    "min",
    "median",
    "mean",
    "std",
    "softplus",
]


def softplus(x: NdarrayOrTensor) -> NdarrayOrTensor:
    """stable softplus through `np.logaddexp` with equivalent implementation for torch.

    Args:
        x: array/tensor.

    Returns:
        Softplus of the input.
    """
    if isinstance(x, np.ndarray):
        return np.logaddexp(np.zeros_like(x), x)
    return torch.logaddexp(torch.zeros_like(x), x)


def allclose(a: NdarrayTensor, b: NdarrayOrTensor, rtol=1e-5, atol=1e-8, equal_nan=False) -> bool:
    """`np.allclose` with equivalent implementation for torch."""
    b, *_ = convert_to_dst_type(b, a, wrap_sequence=True)
    if isinstance(a, np.ndarray):
        return np.allclose(a, b, rtol=rtol, atol=atol, equal_nan=equal_nan)
    return torch.allclose(a, b, rtol=rtol, atol=atol, equal_nan=equal_nan)  # type: ignore


def moveaxis(x: NdarrayOrTensor, src: int | Sequence[int], dst: int | Sequence[int]) -> NdarrayOrTensor:
    """`moveaxis` for pytorch and numpy"""
    if isinstance(x, torch.Tensor):
        return torch.movedim(x, src, dst)  # type: ignore
    return np.moveaxis(x, src, dst)


def in1d(x, y):
    """`np.in1d` with equivalent implementation for torch."""
    if isinstance(x, np.ndarray):
        return np.isin(x, y)
    return (x[..., None] == torch.tensor(y, device=x.device)).any(-1).view(-1)


def clip(a: NdarrayOrTensor, a_min, a_max) -> NdarrayOrTensor:
    """`np.clip` with equivalent implementation for torch."""
    result: NdarrayOrTensor
    if isinstance(a, np.ndarray):
        result = np.clip(a, a_min, a_max)
    else:
        result = torch.clamp(a, a_min, a_max)
    return result


def percentile(
    x: NdarrayOrTensor, q, dim: int | None = None, keepdim: bool = False, **kwargs
) -> NdarrayOrTensor | float | int:
    """`np.percentile` with equivalent implementation for torch.

    Pytorch uses `quantile`. For more details please refer to:
    https://pytorch.org/docs/stable/generated/torch.quantile.html.
    https://numpy.org/doc/stable/reference/generated/numpy.percentile.html.

    Args:
        x: input data.
        q: percentile to compute (should in range 0 <= q <= 100).
        dim: the dim along which the percentiles are computed. default is to compute the percentile
            along a flattened version of the array.
        keepdim: whether the output data has dim retained or not.
        kwargs: if `x` is numpy array, additional args for `np.percentile`, more details:
            https://numpy.org/doc/stable/reference/generated/numpy.percentile.html.

    Returns:
        Resulting value (scalar)
    """
    q_np = convert_data_type(q, output_type=np.ndarray, wrap_sequence=True)[0]
    if ((q_np < 0) | (q_np > 100)).any():
        raise ValueError(f"q values must be in [0, 100], got values: {q}.")
    result: NdarrayOrTensor | float | int
    if isinstance(x, np.ndarray) or (isinstance(x, torch.Tensor) and torch.numel(x) > 1_000_000):  # pytorch#64947
        _x = convert_data_type(x, output_type=np.ndarray)[0]
        result = np.percentile(_x, q_np, axis=dim, keepdims=keepdim, **kwargs)
        result = convert_to_dst_type(result, x)[0]
    else:
        q = convert_to_dst_type(q_np / 100.0, x)[0]
        result = torch.quantile(x, q, dim=dim, keepdim=keepdim)
    return result


def where(condition: NdarrayOrTensor, x=None, y=None) -> NdarrayOrTensor:
    """
    Note that `torch.where` may convert y.dtype to x.dtype.
    """
    result: NdarrayOrTensor
    if isinstance(condition, np.ndarray):
        if x is not None:
            result = np.where(condition, x, y)
        else:
            result = np.where(condition)  # type: ignore
    else:
        if x is not None:
            x = torch.as_tensor(x, device=condition.device)
            y = torch.as_tensor(y, device=condition.device, dtype=x.dtype)
            result = torch.where(condition, x, y)
        else:
            result = torch.where(condition)  # type: ignore
    return result


def argwhere(a: NdarrayTensor) -> NdarrayTensor:
    """`np.argwhere` with equivalent implementation for torch.

    Args:
        a: input data.

    Returns:
        Indices of elements that are non-zero. Indices are grouped by element.
        This array will have shape (N, a.ndim) where N is the number of non-zero items.
    """
    if isinstance(a, np.ndarray):
        return np.argwhere(a)  # type: ignore
    return torch.argwhere(a)  # type: ignore


def argsort(a: NdarrayTensor, axis: int | None = -1) -> NdarrayTensor:
    """`np.argsort` with equivalent implementation for torch.

    Args:
        a: the array/tensor to sort.
        axis: axis along which to sort.

    Returns:
        Array/Tensor of indices that sort a along the specified axis.
    """
    if isinstance(a, np.ndarray):
        return np.argsort(a, axis=axis)  # type: ignore
    return torch.argsort(a, dim=axis)  # type: ignore


def nonzero(x: NdarrayOrTensor) -> NdarrayOrTensor:
    """`np.nonzero` with equivalent implementation for torch.

    Args:
        x: array/tensor.

    Returns:
        Index unravelled for given shape
    """
    if isinstance(x, np.ndarray):
        return np.nonzero(x)[0]
    return torch.nonzero(x).flatten()


def floor_divide(a: NdarrayOrTensor, b) -> NdarrayOrTensor:
    """`np.floor_divide` with equivalent implementation for torch.

    As of pt1.8, use `torch.div(..., rounding_mode="floor")`, and
    before that, use `torch.floor_divide`.

    Args:
        a: first array/tensor
        b: scalar to divide by

    Returns:
        Element-wise floor division between two arrays/tensors.
    """
    if isinstance(a, torch.Tensor):
        return torch.floor_divide(a, b)
    else:
        return np.floor_divide(a, b)


def unravel_index(idx, shape) -> NdarrayOrTensor:
    """`np.unravel_index` with equivalent implementation for torch.

    Args:
        idx: index to unravel.
        shape: shape of array/tensor.

    Returns:
        Index unravelled for given shape
    """
    if isinstance(idx, torch.Tensor):
        coord = []
        for dim in reversed(shape):
            coord.append(idx % dim)
            idx = floor_divide(idx, dim)
        return torch.stack(coord[::-1])
    return np.asarray(np.unravel_index(idx, shape))


def unravel_indices(idx, shape) -> NdarrayOrTensor:
    """Computing unravel coordinates from indices.

    Args:
        idx: a sequence of indices to unravel.
        shape: shape of array/tensor.

    Returns:
        Stacked indices unravelled for given shape
    """
    lib_stack = torch.stack if isinstance(idx[0], torch.Tensor) else np.stack
    return lib_stack([unravel_index(i, shape) for i in idx])  # type: ignore


def ravel(x: NdarrayOrTensor) -> NdarrayOrTensor:
    """`np.ravel` with equivalent implementation for torch.

    Args:
        x: array/tensor to ravel.

    Returns:
        Return a contiguous flattened array/tensor.
    """
    if isinstance(x, torch.Tensor):
        if hasattr(torch, "ravel"):  # `ravel` is new in torch 1.8.0
            return x.ravel()
        return x.flatten().contiguous()
    return np.ravel(x)


def any_np_pt(x: NdarrayOrTensor, axis: int | Sequence[int]) -> NdarrayOrTensor:
    """`np.any` with equivalent implementation for torch.

    For pytorch, convert to boolean for compatibility with older versions.

    Args:
        x: input array/tensor.
        axis: axis to perform `any` over.

    Returns:
        Return a contiguous flattened array/tensor.
    """
    if isinstance(x, np.ndarray):
        return np.any(x, axis)  # type: ignore

    # pytorch can't handle multiple dimensions to `any` so loop across them
    axis = [axis] if not isinstance(axis, Sequence) else axis
    for ax in axis:
        try:
            x = torch.any(x, ax)
        except RuntimeError:
            # older versions of pytorch require the input to be cast to boolean
            x = torch.any(x.bool(), ax)
    return x


def maximum(a: NdarrayOrTensor, b: NdarrayOrTensor) -> NdarrayOrTensor:
    """`np.maximum` with equivalent implementation for torch.

    Args:
        a: first array/tensor.
        b: second array/tensor.

    Returns:
        Element-wise maximum between two arrays/tensors.
    """
    if isinstance(a, torch.Tensor) and isinstance(b, torch.Tensor):
        return torch.maximum(a, b)
    return np.maximum(a, b)


def concatenate(to_cat: Sequence[NdarrayOrTensor], axis: int = 0, out=None) -> NdarrayOrTensor:
    """`np.concatenate` with equivalent implementation for torch (`torch.cat`)."""
    if isinstance(to_cat[0], np.ndarray):
        return np.concatenate(to_cat, axis, out)  # type: ignore
    return torch.cat(to_cat, dim=axis, out=out)  # type: ignore


def cumsum(a: NdarrayOrTensor, axis=None, **kwargs) -> NdarrayOrTensor:
    """
    `np.cumsum` with equivalent implementation for torch.

    Args:
        a: input data to compute cumsum.
        axis: expected axis to compute cumsum.
        kwargs: if `a` is PyTorch Tensor, additional args for `torch.cumsum`, more details:
            https://pytorch.org/docs/stable/generated/torch.cumsum.html.

    """

    if isinstance(a, np.ndarray):
        return np.cumsum(a, axis)  # type: ignore
    if axis is None:
        return torch.cumsum(a[:], 0, **kwargs)
    return torch.cumsum(a, dim=axis, **kwargs)


def isfinite(x: NdarrayOrTensor) -> NdarrayOrTensor:
    """`np.isfinite` with equivalent implementation for torch."""
    if not isinstance(x, torch.Tensor):
        return np.isfinite(x)  # type: ignore
    return torch.isfinite(x)


def searchsorted(a: NdarrayTensor, v: NdarrayOrTensor, right=False, sorter=None, **kwargs) -> NdarrayTensor:
    """
    `np.searchsorted` with equivalent implementation for torch.

    Args:
        a: numpy array or tensor, containing monotonically increasing sequence on the innermost dimension.
        v: containing the search values.
        right: if False, return the first suitable location that is found, if True, return the last such index.
        sorter: if `a` is numpy array, optional array of integer indices that sort array `a` into ascending order.
        kwargs: if `a` is PyTorch Tensor, additional args for `torch.searchsorted`, more details:
            https://pytorch.org/docs/stable/generated/torch.searchsorted.html.

    """
    side = "right" if right else "left"
    if isinstance(a, np.ndarray):
        return np.searchsorted(a, v, side, sorter)  # type: ignore
    return torch.searchsorted(a, v, right=right, **kwargs)  # type: ignore


def repeat(a: NdarrayOrTensor, repeats: int, axis: int | None = None, **kwargs) -> NdarrayOrTensor:
    """
    `np.repeat` with equivalent implementation for torch (`repeat_interleave`).

    Args:
        a: input data to repeat.
        repeats: number of repetitions for each element, repeats is broadcast to fit the shape of the given axis.
        axis: axis along which to repeat values.
        kwargs: if `a` is PyTorch Tensor, additional args for `torch.repeat_interleave`, more details:
            https://pytorch.org/docs/stable/generated/torch.repeat_interleave.html.

    """
    if isinstance(a, np.ndarray):
        return np.repeat(a, repeats, axis)
    return torch.repeat_interleave(a, repeats, dim=axis, **kwargs)


def isnan(x: NdarrayOrTensor) -> NdarrayOrTensor:
    """`np.isnan` with equivalent implementation for torch.

    Args:
        x: array/tensor.

    """
    if isinstance(x, np.ndarray):
        return np.isnan(x)  # type: ignore
    return torch.isnan(x)


T = TypeVar("T")


def ascontiguousarray(x: NdarrayTensor | T, **kwargs) -> NdarrayOrTensor | T:
    """`np.ascontiguousarray` with equivalent implementation for torch (`contiguous`).

    Args:
        x: array/tensor.
        kwargs: if `x` is PyTorch Tensor, additional args for `torch.contiguous`, more details:
            https://pytorch.org/docs/stable/generated/torch.Tensor.contiguous.html.

    """
    if isinstance(x, np.ndarray):
        if x.ndim == 0:
            return x
        return np.ascontiguousarray(x)
    if isinstance(x, torch.Tensor):
        return x.contiguous(**kwargs)
    return x


def stack(x: Sequence[NdarrayTensor], dim: int) -> NdarrayTensor:
    """`np.stack` with equivalent implementation for torch.

    Args:
        x: array/tensor.
        dim: dimension along which to perform the stack (referred to as `axis` by numpy).
    """
    if isinstance(x[0], np.ndarray):
        return np.stack(x, dim)  # type: ignore
    return torch.stack(x, dim)  # type: ignore


def mode(x: NdarrayTensor, dim: int = -1, to_long: bool = True) -> NdarrayTensor:
    """`torch.mode` with equivalent implementation for numpy.

    Args:
        x: array/tensor.
        dim: dimension along which to perform `mode` (referred to as `axis` by numpy).
        to_long: convert input to long before performing mode.
    """
    dtype = torch.int64 if to_long else None
    x_t, *_ = convert_data_type(x, torch.Tensor, dtype=dtype)
    o_t = torch.mode(x_t, dim).values
    o, *_ = convert_to_dst_type(o_t, x)
    return o


def unique(x: NdarrayTensor, **kwargs) -> NdarrayTensor:
    """`torch.unique` with equivalent implementation for numpy.

    Args:
        x: array/tensor.
    """
    return np.unique(x, **kwargs) if isinstance(x, (np.ndarray, list)) else torch.unique(x, **kwargs)  # type: ignore


def linalg_inv(x: NdarrayTensor) -> NdarrayTensor:
    """`torch.linalg.inv` with equivalent implementation for numpy.

    Args:
        x: array/tensor.
    """
    if isinstance(x, torch.Tensor) and hasattr(torch, "inverse"):  # pytorch 1.7.0
        return torch.inverse(x)  # type: ignore
    return torch.linalg.inv(x) if isinstance(x, torch.Tensor) else np.linalg.inv(x)  # type: ignore


def max(x: NdarrayTensor, dim: int | tuple | None = None, **kwargs) -> NdarrayTensor:
    """`torch.max` with equivalent implementation for numpy

    Args:
        x: array/tensor.

    Returns:
        the maximum of x.

    """

    ret: NdarrayTensor
    if dim is None:
        ret = np.max(x, **kwargs) if isinstance(x, (np.ndarray, list)) else torch.max(x, **kwargs)  # type: ignore
    else:
        if isinstance(x, (np.ndarray, list)):
            ret = np.max(x, axis=dim, **kwargs)
        else:
            ret = torch.max(x, int(dim), **kwargs)  # type: ignore

    return ret[0] if isinstance(ret, tuple) else ret


def mean(x: NdarrayTensor, dim: int | tuple | None = None, **kwargs) -> NdarrayTensor:
    """`torch.mean` with equivalent implementation for numpy

    Args:
        x: array/tensor.

    Returns:
        the mean of x
    """

    ret: NdarrayTensor
    if dim is None:
        ret = np.mean(x, **kwargs) if isinstance(x, (np.ndarray, list)) else torch.mean(x, **kwargs)  # type: ignore
    else:
        if isinstance(x, (np.ndarray, list)):
            ret = np.mean(x, axis=dim, **kwargs)
        else:
            ret = torch.mean(x, int(dim), **kwargs)  # type: ignore

    return ret


def median(x: NdarrayTensor, dim: int | tuple | None = None, **kwargs) -> NdarrayTensor:
    """`torch.median` with equivalent implementation for numpy

    Args:
        x: array/tensor.

    Returns
        the median of x.
    """

    ret: NdarrayTensor
    if dim is None:
        ret = np.median(x, **kwargs) if isinstance(x, (np.ndarray, list)) else torch.median(x, **kwargs)  # type: ignore
    else:
        if isinstance(x, (np.ndarray, list)):
            ret = np.median(x, axis=dim, **kwargs)
        else:
            ret = torch.median(x, int(dim), **kwargs)  # type: ignore

    return ret


def min(x: NdarrayTensor, dim: int | tuple | None = None, **kwargs) -> NdarrayTensor:
    """`torch.min` with equivalent implementation for numpy

    Args:
        x: array/tensor.

    Returns:
        the minimum of x.
    """

    ret: NdarrayTensor
    if dim is None:
        ret = np.min(x, **kwargs) if isinstance(x, (np.ndarray, list)) else torch.min(x, **kwargs)  # type: ignore
    else:
        if isinstance(x, (np.ndarray, list)):
            ret = np.min(x, axis=dim, **kwargs)
        else:
            ret = torch.min(x, int(dim), **kwargs)  # type: ignore

    return ret[0] if isinstance(ret, tuple) else ret


def std(x: NdarrayTensor, dim: int | tuple | None = None, unbiased: bool = False) -> NdarrayTensor:
    """`torch.std` with equivalent implementation for numpy

    Args:
        x: array/tensor.

    Returns:
        the standard deviation of x.
    """

    ret: NdarrayTensor
    if dim is None:
        ret = np.std(x) if isinstance(x, (np.ndarray, list)) else torch.std(x, unbiased)  # type: ignore
    else:
        if isinstance(x, (np.ndarray, list)):
            ret = np.std(x, axis=dim)
        else:
            ret = torch.std(x, int(dim), unbiased)  # type: ignore

    return ret


def sum(x: NdarrayTensor, dim: int | tuple | None = None, **kwargs) -> NdarrayTensor:
    """`torch.sum` with equivalent implementation for numpy

    Args:
        x: array/tensor.

    Returns:
        the sum of x.
    """

    ret: NdarrayTensor
    if dim is None:
        ret = np.sum(x, **kwargs) if isinstance(x, (np.ndarray, list)) else torch.sum(x, **kwargs)  # type: ignore
    else:
        if isinstance(x, (np.ndarray, list)):
            ret = np.sum(x, axis=dim, **kwargs)
        else:
            ret = torch.sum(x, int(dim), **kwargs)  # type: ignore

    return ret



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
from parameterized import parameterized

from monai.apps.nuclick.transforms import (
    AddClickSignalsd,
    AddLabelAsGuidanced,
    AddPointGuidanceSignald,
    ExtractPatchd,
    FilterImaged,
    FlattenLabeld,
    PostFilterLabeld,
    SetLabelClassd,
    SplitLabeld,
)

# Data Definitions
RGB_IMAGE_1 = np.array(
    [[[0, 0, 0], [0, 1, 0], [0, 0, 1]], [[2, 0, 2], [0, 1, 0], [1, 0, 1]], [[3, 0, 2], [0, 1, 0], [1, 3, 1]]]
)

LABEL_1 = np.array(
    [
        [1, 1, 1, 0, 0, 0, 0],
        [1, 1, 1, 0, 0, 0, 0],
        [1, 1, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0],
        [1, 1, 1, 0, 1, 1, 1],
        [1, 1, 1, 0, 1, 1, 1],
        [1, 1, 1, 0, 1, 1, 1],
    ],
    dtype=np.uint8,
)

LABEL_1_1 = np.array(
    [
        [1, 1, 1, 0, 0, 1, 1],
        [1, 1, 1, 0, 0, 1, 1],
        [1, 1, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0],
        [1, 1, 1, 0, 2, 2, 2],
        [1, 1, 1, 0, 2, 2, 2],
        [1, 1, 1, 0, 2, 2, 2],
    ],
    dtype=np.uint8,
)

LABEL_2 = np.array([[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]], dtype=np.uint8)

LABEL_3 = np.array([[[1, 1, 1, 1], [2, 2, 2, 2], [3, 3, 3, 3], [4, 4, 4, 4]]], dtype=np.uint8)

LABEL_4 = np.array([[[4, 4, 4, 4], [4, 4, 4, 4], [4, 4, 4, 4], [4, 4, 4, 4]]], dtype=np.uint8)

IL_IMAGE_1 = np.array(
    [
        [[0, 0, 0, 0, 0], [0, 1, 0, 0, 0], [0, 0, 1, 1, 1], [0, 0, 1, 1, 1], [0, 0, 1, 1, 1]],
        [[0, 0, 0, 0, 0], [0, 1, 0, 0, 0], [0, 0, 1, 1, 1], [0, 0, 1, 1, 1], [0, 0, 1, 1, 1]],
        [[0, 0, 0, 0, 0], [0, 1, 0, 0, 0], [0, 0, 1, 1, 1], [0, 0, 1, 1, 1], [0, 0, 1, 1, 1]],
    ]
)

IL_FG_IMAGE_1 = np.array([[0, 0, 0, 0, 0], [0, 1, 0, 0, 0], [0, 0, 1, 1, 1], [0, 0, 1, 1, 1], [0, 0, 1, 1, 1]])

IL_LABEL_1 = np.array(
    [[[0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 1, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0]]], dtype=np.uint8
)

IL_OTHERS_1 = np.array(
    [[[1, 1, 1, 1, 1], [2, 0, 0, 0, 2], [3, 0, 0, 0, 3], [4, 0, 0, 0, 4], [5, 5, 5, 5, 5]]], dtype=np.uint8
)

IL_IMAGE_2 = np.array(
    [[[0, 0, 0], [0, 1, 0], [0, 0, 1]], [[0, 0, 0], [0, 1, 0], [0, 0, 1]], [[0, 0, 0], [0, 1, 0], [0, 0, 1]]]
)

IL_LABEL_2 = np.array([[[0, 0, 0], [0, 1, 0], [0, 0, 0]]], dtype=np.uint8)

PRED_1 = np.array(
    [[[1, 1, 1, 1, 1], [2, 0, 0, 0, 2], [3, 0, 0, 0, 3], [4, 0, 0, 0, 4], [5, 5, 5, 5, 5]]], dtype=np.float32
)

NUC_POINTS_1 = np.array(
    [
        [[[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 1, 0], [0, 0, 0, 0]]],
        [[[0, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]]],
    ],
    dtype=np.float32,
)
BB_1 = np.array([[0, 0, 5, 5], [1, 1, 6, 6]], dtype=np.uint8)

DATA_FILTER_1 = {"image": RGB_IMAGE_1}

DATA_FLATTEN_1 = {"label": LABEL_1}
DATA_FLATTEN_2 = {"label": LABEL_2}

DATA_EXTRACT_1 = {"image": IL_IMAGE_1, "label": IL_LABEL_1, "centroid": (2, 2)}
DATA_EXTRACT_2 = {"image": IL_IMAGE_2, "label": IL_LABEL_2, "centroid": (1, 1)}

DATA_SPLIT_1 = {"label": LABEL_3, "mask_value": 1}
DATA_SPLIT_2 = {"label": LABEL_4, "mask_value": 4}

DATA_GUIDANCE_1 = {"image": IL_IMAGE_1, "label": IL_LABEL_1, "others": IL_OTHERS_1, "centroid": (2, 2)}

DATA_CLICK_1 = {"image": IL_IMAGE_1, "foreground": [[2, 2], [1, 1]]}

DATA_LABEL_FILTER_1 = {
    "pred": PRED_1,
    "nuc_points": NUC_POINTS_1,
    "bounding_boxes": BB_1,
    "img_height": 6,
    "img_width": 6,
}

# Result Definitions
EXTRACT_RESULT_TC1 = np.array([[[0, 0, 0], [0, 0, 0], [0, 0, 1]]], dtype=np.uint8)
EXTRACT_RESULT_TC2 = np.array([[[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]]], dtype=np.uint8)

SPLIT_RESULT_TC1 = np.array([[[1, 1, 1, 1], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]]], dtype=np.uint8)
SPLIT_RESULT_TC2 = np.array([[[1, 1, 1, 1], [1, 1, 1, 1], [1, 1, 1, 1], [1, 1, 1, 1]]], dtype=np.uint8)

# Test Case Definitions
FILTER_IMAGE_TEST_CASE_1 = [{"keys": "image", "min_size": 1}, DATA_FILTER_1, [3, 3, 3]]

FLATTEN_LABEL_TEST_CASE_1 = [{"keys": "label"}, DATA_FLATTEN_1, [0, 1, 2, 3]]
FLATTEN_LABEL_TEST_CASE_2 = [{"keys": "label"}, DATA_FLATTEN_2, [0]]
FLATTEN_LABEL_TEST_CASE_3 = [{"keys": "label"}, {"label": LABEL_1_1}, [0, 1, 2, 3, 4]]

EXTRACT_TEST_CASE_1 = [{"keys": ["image", "label"], "patch_size": 3}, DATA_EXTRACT_1, [1, 3, 3]]
EXTRACT_TEST_CASE_2 = [{"keys": ["image", "label"], "patch_size": 5}, DATA_EXTRACT_1, [1, 5, 5]]
EXTRACT_TEST_CASE_3 = [{"keys": ["image", "label"], "patch_size": 1}, DATA_EXTRACT_2, [1, 1, 1]]

EXTRACT_RESULT_TEST_CASE_1 = [{"keys": ["image", "label"], "patch_size": 3}, DATA_EXTRACT_1, EXTRACT_RESULT_TC1]
EXTRACT_RESULT_TEST_CASE_2 = [{"keys": ["image", "label"], "patch_size": 4}, DATA_EXTRACT_2, EXTRACT_RESULT_TC2]

EXTRACT_KW_TEST_CASE_1 = [
    {"keys": ["image", "label"], "patch_size": 3, "mode": "constant"},
    DATA_EXTRACT_1,
    EXTRACT_RESULT_TC1,
]

SPLIT_TEST_CASE_1 = [{"keys": ["label"], "mask_value": "mask_value", "min_area": 1}, DATA_SPLIT_1, SPLIT_RESULT_TC1]
SPLIT_TEST_CASE_2 = [{"keys": ["label"], "mask_value": "mask_value", "min_area": 3}, DATA_SPLIT_2, SPLIT_RESULT_TC2]

GUIDANCE_TEST_CASE_1 = [{"image": "image", "label": "label", "others": "others"}, DATA_GUIDANCE_1, [5, 5, 5]]
GUIDANCE_TEST_CASE_2 = [
    {"image": "image", "label": "label", "others": "others", "gaussian": True, "use_distance": True},
    DATA_GUIDANCE_1,
    [5, 5, 5],
]

CLICK_TEST_CASE_1 = [{"image": "image", "foreground": "foreground", "bb_size": 4}, DATA_CLICK_1, [2, 5, 4, 4]]
CLICK_TEST_CASE_2 = [
    {"image": "image", "foreground": "foreground", "bb_size": 4, "gaussian": True},
    DATA_CLICK_1,
    [2, 5, 4, 4],
]

LABEL_FILTER_TEST_CASE_1 = [{"keys": ["pred"]}, DATA_LABEL_FILTER_1, [6, 6]]

LABEL_GUIDANCE_TEST_CASE_1 = [{"keys": ["image"], "source": "label"}, DATA_GUIDANCE_1, [4, 5, 5]]

LABEL_CLASS_TEST_CASE_1 = [{"keys": ["label"], "offset": 2}, DATA_GUIDANCE_1, 3]

# Test Case Classes


class TestFilterImaged(unittest.TestCase):

    @parameterized.expand([FILTER_IMAGE_TEST_CASE_1])
    def test_correct_shape(self, arguments, input_data, expected_shape):
        result = FilterImaged(**arguments)(input_data)
        np.testing.assert_equal(result["image"].shape, expected_shape)


class TestFlattenLabeld(unittest.TestCase):

    @parameterized.expand([FLATTEN_LABEL_TEST_CASE_1, FLATTEN_LABEL_TEST_CASE_2, FLATTEN_LABEL_TEST_CASE_3])
    def test_correct_num_labels(self, arguments, input_data, expected_result):
        result = FlattenLabeld(**arguments)(input_data)
        np.testing.assert_equal(np.unique(result["label"]), expected_result)


class TestExtractPatchd(unittest.TestCase):

    @parameterized.expand([EXTRACT_TEST_CASE_1, EXTRACT_TEST_CASE_2, EXTRACT_TEST_CASE_3])
    def test_correct_patch_size(self, arguments, input_data, expected_shape):
        result = ExtractPatchd(**arguments)(input_data)
        np.testing.assert_equal(result["label"].shape, expected_shape)

    @parameterized.expand([EXTRACT_RESULT_TEST_CASE_1, EXTRACT_RESULT_TEST_CASE_2, EXTRACT_KW_TEST_CASE_1])
    def test_correct_results(self, arguments, input_data, expected_result):
        result = ExtractPatchd(**arguments)(input_data)
        np.testing.assert_equal(result["label"], expected_result)


class TestSplitLabelsd(unittest.TestCase):

    @parameterized.expand([SPLIT_TEST_CASE_1, SPLIT_TEST_CASE_2])
    def test_correct_results(self, arguments, input_data, expected_result):
        result = SplitLabeld(**arguments)(input_data)
        np.testing.assert_equal(result["label"], expected_result)


class TestGuidanceSignal(unittest.TestCase):

    @parameterized.expand([GUIDANCE_TEST_CASE_1, GUIDANCE_TEST_CASE_2])
    def test_correct_shape(self, arguments, input_data, expected_shape):
        result = AddPointGuidanceSignald(**arguments)(input_data)
        np.testing.assert_equal(result["image"].shape, expected_shape)


class TestClickSignal(unittest.TestCase):

    @parameterized.expand([CLICK_TEST_CASE_1, CLICK_TEST_CASE_2])
    def test_correct_shape(self, arguments, input_data, expected_shape):
        result = AddClickSignalsd(**arguments)(input_data)
        np.testing.assert_equal(result["image"].shape, expected_shape)


class TestPostFilterLabel(unittest.TestCase):

    @parameterized.expand([LABEL_FILTER_TEST_CASE_1])
    def test_correct_shape(self, arguments, input_data, expected_shape):
        result = PostFilterLabeld(**arguments)(input_data)
        np.testing.assert_equal(result["pred"].shape, expected_shape)


class TestAddLabelAsGuidance(unittest.TestCase):

    @parameterized.expand([LABEL_GUIDANCE_TEST_CASE_1])
    def test_correct_shape(self, arguments, input_data, expected_shape):
        result = AddLabelAsGuidanced(**arguments)(input_data)
        np.testing.assert_equal(result["image"].shape, expected_shape)


class TestSetLabelClass(unittest.TestCase):

    @parameterized.expand([LABEL_CLASS_TEST_CASE_1])
    def test_correct_results(self, arguments, input_data, expected_result):
        result = SetLabelClassd(**arguments)(input_data)
        assert result["label"] == expected_result


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest
from copy import deepcopy

import numpy as np
import torch
from numpy.fft import fftn, fftshift
from parameterized import parameterized

from monai.data.synthetic import create_test_image_2d, create_test_image_3d
from monai.transforms import KSpaceSpikeNoised
from monai.utils.misc import set_determinism
from tests.test_utils import TEST_NDARRAYS, assert_allclose

TESTS = []
for shape in ((128, 64), (64, 48, 80)):
    for p in TEST_NDARRAYS:
        TESTS.append((shape, p))

KEYS = ["image", "label"]


class TestKSpaceSpikeNoised(unittest.TestCase):
    def setUp(self):
        set_determinism(0)
        super().setUp()

    def tearDown(self):
        set_determinism(None)

    @staticmethod
    def get_data(im_shape, im_type):
        create_test_image = create_test_image_2d if len(im_shape) == 2 else create_test_image_3d
        ims = create_test_image(*im_shape, rad_max=20, noise_max=0.0, num_seg_classes=5)
        ims = [im_type(im[None]) for im in ims]
        return dict(zip(KEYS, ims))

    @parameterized.expand(TESTS)
    def test_same_result(self, im_shape, im_type):
        data = self.get_data(im_shape, im_type)
        loc = [0] + [int(im_shape[i] / 2) for i in range(len(im_shape))]
        k_intensity = 10

        t = KSpaceSpikeNoised(KEYS, loc, k_intensity)
        out1 = t(deepcopy(data))
        out2 = t(deepcopy(data))

        for k in KEYS:
            if isinstance(out1[k], torch.Tensor):
                out1[k] = out1[k].cpu()
                out2[k] = out2[k].cpu()
            np.testing.assert_allclose(out1[k], out2[k])

    @parameterized.expand(TESTS)
    def test_highlighted_kspace_pixel(self, im_shape, im_type):
        data = self.get_data(im_shape, im_type)
        loc = [0] + [int(im_shape[i] / 2) for i in range(len(im_shape))]
        k_intensity = 10

        t = KSpaceSpikeNoised(KEYS, loc, k_intensity)
        out = t(data)

        for k in KEYS:
            if isinstance(out[k], torch.Tensor):
                out[k] = out[k].cpu()

            n_dims = len(im_shape)
            out_k = fftshift(fftn(out[k], axes=tuple(range(-n_dims, 0))), axes=tuple(range(-n_dims, 0)))
            log_mag = np.log(np.absolute(out_k))
            np.testing.assert_allclose(k_intensity, log_mag[tuple(loc)], 1e-1)

    @parameterized.expand(TESTS)
    def test_dict_matches(self, im_shape, im_type):
        data = self.get_data(im_shape, im_type)
        # use same image for both dictionary entries to check same trans is applied to them
        data = {KEYS[0]: deepcopy(data[KEYS[0]]), KEYS[1]: deepcopy(data[KEYS[0]])}
        loc = [0] + [int(im_shape[i] / 2) for i in range(len(im_shape))]
        k_intensity = 10

        t = KSpaceSpikeNoised(KEYS, loc, k_intensity)
        out = t(deepcopy(data))
        assert_allclose(out[KEYS[0]], out[KEYS[1]], type_test=False)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections.abc import Callable

from monai.handlers.ignite_metric import IgniteMetricHandler
from monai.metrics import HausdorffDistanceMetric
from monai.utils import MetricReduction


class HausdorffDistance(IgniteMetricHandler):
    """
    Computes Hausdorff distance from full size Tensor and collects average over batch, class-channels, iterations.
    """

    def __init__(
        self,
        include_background: bool = False,
        distance_metric: str = "euclidean",
        percentile: float | None = None,
        directed: bool = False,
        reduction: MetricReduction | str = MetricReduction.MEAN,
        output_transform: Callable = lambda x: x,
        save_details: bool = True,
    ) -> None:
        """

        Args:
            include_background: whether to include distance computation on the first channel of the predicted output.
                Defaults to ``False``.
            distance_metric: : [``"euclidean"``, ``"chessboard"``, ``"taxicab"``]
                the metric used to compute surface distance. Defaults to ``"euclidean"``.
            percentile: an optional float number between 0 and 100. If specified, the corresponding
                percentile of the Hausdorff Distance rather than the maximum result will be achieved.
                Defaults to ``None``.
            directed: whether to calculate directed Hausdorff distance. Defaults to ``False``.
            reduction: define the mode to reduce metrics, will only execute reduction on `not-nan` values,
                available reduction modes: {``"none"``, ``"mean"``, ``"sum"``, ``"mean_batch"``, ``"sum_batch"``,
                ``"mean_channel"``, ``"sum_channel"``}, default to ``"mean"``. if "none", will not do reduction.
            output_transform: callable to extract `y_pred` and `y` from `ignite.engine.state.output` then
                construct `(y_pred, y)` pair, where `y_pred` and `y` can be `batch-first` Tensors or
                lists of `channel-first` Tensors. the form of `(y_pred, y)` is required by the `update()`.
                `engine.state` and `output_transform` inherit from the ignite concept:
                https://pytorch.org/ignite/concepts.html#state, explanation and usage example are in the tutorial:
                https://github.com/Project-MONAI/tutorials/blob/master/modules/batch_output_transform.ipynb.
            save_details: whether to save metric computation details per image, for example: hausdorff distance
                of every image. default to True, will save to `engine.state.metric_details` dict with the metric name as key.

        """
        metric_fn = HausdorffDistanceMetric(
            include_background=include_background,
            distance_metric=distance_metric,
            percentile=percentile,
            directed=directed,
            reduction=reduction,
        )
        super().__init__(metric_fn=metric_fn, output_transform=output_transform, save_details=save_details)



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
from collections.abc import Sequence

import numpy as np
from numpy import ndarray

from monai.config import PathLike
from monai.data.image_reader import ImageReader
from monai.data.utils import is_supported_format
from monai.utils import FastMRIKeys, optional_import, require_pkg

h5py, has_h5py = optional_import("h5py")


@require_pkg(pkg_name="h5py")
class FastMRIReader(ImageReader):
    """
    Load fastMRI files with '.h5' suffix. fastMRI files, when loaded with "h5py",
    are HDF5 dictionary-like datasets. The keys are:

    - kspace: contains the fully-sampled kspace
    - reconstruction_rss: contains the root sum of squares of ifft of kspace. This
        is the ground-truth image.

    It also has several attributes with the following keys:

    - acquisition (str): acquisition mode of the data (e.g., AXT2 denotes T2 brain MRI scans)
    - max (float): dynamic range of the data
    - norm (float): norm of the kspace
    - patient_id (str): the patient's id whose measurements were recorded
    """

    def verify_suffix(self, filename: Sequence[PathLike] | PathLike) -> bool:
        """
         Verify whether the specified file format is supported by h5py reader.

        Args:
             filename: file name
        """
        suffixes: Sequence[str] = [".h5"]
        return has_h5py and is_supported_format(filename, suffixes)

    def read(self, data: Sequence[PathLike] | PathLike) -> dict:  # type: ignore
        """
        Read data from specified h5 file.
        Note that the returned object is a dictionary.

        Args:
            data: file name to read.
        """
        if isinstance(data, (tuple, list)):
            data = data[0]

        with h5py.File(data, "r") as f:
            # extract everything from the ht5 file
            dat = dict(
                [(key, f[key][()]) for key in f]
                + [(key, f.attrs[key]) for key in f.attrs]
                + [(FastMRIKeys.FILENAME, os.path.basename(data))]  # type: ignore
            )
        f.close()

        return dat

    def get_data(self, dat: dict) -> tuple[ndarray, dict]:
        """
        Extract data array and metadata from the loaded data and return them.
        This function returns two objects, first is numpy array of image data, second is dict of metadata.

        Args:
            dat: a dictionary loaded from an h5 file
        """
        header = self._get_meta_dict(dat)
        data: ndarray = np.array(dat[FastMRIKeys.KSPACE])
        header[FastMRIKeys.MASK] = (
            np.expand_dims(np.array(dat[FastMRIKeys.MASK]), 0)[None, ..., None]
            if FastMRIKeys.MASK in dat.keys()
            else np.zeros(data.shape)
        )
        return data, header

    def _get_meta_dict(self, dat: dict) -> dict:
        """
        Get all the metadata of the loaded dict and return the meta dict.

        Args:
            dat: a dictionary object loaded from an h5 file.
        """
        return {k.value: dat[k.value] for k in FastMRIKeys if k.value in dat}



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest
from copy import deepcopy

import numpy as np
from parameterized import parameterized

from monai.data.meta_tensor import MetaTensor
from monai.transforms.transform import Randomizable
from monai.transforms.utility.array import RandLambda
from tests.test_utils import TEST_NDARRAYS, assert_allclose


class RandTest(Randomizable):
    """
    randomisable transform for testing.
    """

    def randomize(self, data=None):
        self._a = self.R.random()

    def __call__(self, data):
        self.randomize()
        return deepcopy(data) + self._a


class TestRandLambda(unittest.TestCase):
    def check(self, tr: RandLambda, img, img_orig_type, out, expected=None):
        # input shouldn't change
        self.assertIsInstance(img, img_orig_type)
        if isinstance(img, MetaTensor):
            self.assertEqual(len(img.applied_operations), 0)
        # output data matches expected
        assert_allclose(expected, out, type_test=False)
        # output type is MetaTensor with 1 appended operation
        self.assertIsInstance(out, MetaTensor)
        self.assertEqual(len(out.applied_operations), 1)

        # inverse
        inv = tr.inverse(out)
        # after inverse, input image remains unchanged
        self.assertIsInstance(img, img_orig_type)
        if isinstance(img, MetaTensor):
            self.assertEqual(len(img.applied_operations), 0)
        # after inverse, output is MetaTensor with 0 applied operations
        self.assertIsInstance(inv, MetaTensor)
        self.assertEqual(len(inv.applied_operations), 0)

    @parameterized.expand([[p] for p in TEST_NDARRAYS])
    def test_rand_lambdad_identity(self, t):
        img = t(np.zeros((10, 10)))
        img_t = type(img)

        test_func = RandTest()
        test_func.set_random_state(seed=134)
        expected = test_func(img)
        test_func.set_random_state(seed=134)

        # default prob
        tr = RandLambda(func=test_func)
        ret = tr(img)
        self.check(tr, img, img_t, ret, expected)

        tr = RandLambda(func=test_func, prob=0.0)
        ret = tr(img)
        self.check(tr, img, img_t, ret, expected=img)

        trans = RandLambda(func=test_func, prob=0.5)
        trans.set_random_state(seed=123)
        ret = trans(img)
        self.check(trans, img, img_t, ret, expected=img)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING

from monai.utils import IgniteInfo, min_version, optional_import

Events, _ = optional_import("ignite.engine", IgniteInfo.OPT_IMPORT_VERSION, min_version, "Events")
EarlyStopping, _ = optional_import("ignite.handlers", IgniteInfo.OPT_IMPORT_VERSION, min_version, "EarlyStopping")

if TYPE_CHECKING:
    from ignite.engine import Engine
else:
    Engine, _ = optional_import(
        "ignite.engine", IgniteInfo.OPT_IMPORT_VERSION, min_version, "Engine", as_type="decorator"
    )


class EarlyStopHandler:
    """
    EarlyStopHandler acts as an Ignite handler to stop training if no improvement after a given number of events.
    It‘s based on the `EarlyStopping` handler in ignite.

    Args:
        patience: number of events to wait if no improvement and then stop the training.
        score_function: It should be a function taking a single argument, an :class:`~ignite.engine.engine.Engine`
            object that the handler attached, can be a trainer or validator, and return a score `float`.
            an improvement is considered if the score is higher.
        trainer: trainer engine to stop the run if no improvement, if None, must call `set_trainer()` before training.
        min_delta: a minimum increase in the score to qualify as an improvement,
            i.e. an increase of less than or equal to `min_delta`, will count as no improvement.
        cumulative_delta: if True, `min_delta` defines an increase since the last `patience` reset, otherwise,
            it defines an increase after the last event, default to False.
        epoch_level: check early stopping for every epoch or every iteration of the attached engine,
            `True` is epoch level, `False` is iteration level, default to epoch level.

    Note:
        If in distributed training and uses loss value of every iteration to detect early stopping,
        the values may be different in different ranks. When using this handler with distributed training,
        please also note that to prevent "dist.destroy_process_group()" hangs, you can use an "all_reduce" operation
        to synchronize the stop signal across all ranks. The mechanism can be implemented in the `score_function`. The following
        is an example:

        .. code-block:: python

            import os

            import torch
            import torch.distributed as dist


            def score_function(engine):
                val_metric = engine.state.metrics["val_mean_dice"]
                if dist.is_initialized():
                    device = torch.device("cuda:" + os.environ["LOCAL_RANK"])
                    val_metric = torch.tensor([val_metric]).to(device)
                    dist.all_reduce(val_metric, op=dist.ReduceOp.SUM)
                    val_metric /= dist.get_world_size()
                    return val_metric.item()
                return val_metric


        User may attach this handler to validator engine to detect validation metrics and stop the training,
        in this case, the `score_function` is executed on validator engine and `trainer` is the trainer engine.

    """

    def __init__(
        self,
        patience: int,
        score_function: Callable,
        trainer: Engine | None = None,
        min_delta: float = 0.0,
        cumulative_delta: bool = False,
        epoch_level: bool = True,
    ) -> None:
        self.patience = patience
        self.score_function = score_function
        self.min_delta = min_delta
        self.cumulative_delta = cumulative_delta
        self.epoch_level = epoch_level
        self._handler = None

        if trainer is not None:
            self.set_trainer(trainer=trainer)

    def attach(self, engine: Engine) -> None:
        """
        Args:
            engine: Ignite Engine, it can be a trainer, validator or evaluator.
        """
        if self.epoch_level:
            engine.add_event_handler(Events.EPOCH_COMPLETED, self)
        else:
            engine.add_event_handler(Events.ITERATION_COMPLETED, self)

    def set_trainer(self, trainer: Engine) -> None:
        """
        Set trainer to execute early stop if not setting properly in `__init__()`.
        """
        self._handler = EarlyStopping(
            patience=self.patience,
            score_function=self.score_function,
            trainer=trainer,
            min_delta=self.min_delta,
            cumulative_delta=self.cumulative_delta,
        )

    def __call__(self, engine: Engine) -> None:
        if self._handler is None:
            raise RuntimeError("please set trainer in __init__() or call set_trainer() before training.")
        self._handler(engine)



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import gc
from typing import TYPE_CHECKING

from monai.utils import IgniteInfo, min_version, optional_import

if TYPE_CHECKING:
    from ignite.engine import Engine, Events
    from ignite.engine.events import CallableEventWithFilter
else:
    CallableEventWithFilter, _ = optional_import(
        "ignite.engine.events", IgniteInfo.OPT_IMPORT_VERSION, min_version, "CallableEventWithFilter"
    )
    Engine, _ = optional_import("ignite.engine", IgniteInfo.OPT_IMPORT_VERSION, min_version, "Engine")
    Events, _ = optional_import("ignite.engine", IgniteInfo.OPT_IMPORT_VERSION, min_version, "Events")


class GarbageCollector:
    """
    Run garbage collector after each epoch

    Args:
        trigger_event: the event that trigger a call to this handler.
            - "epoch", after completion of each epoch (equivalent of ignite.engine.Events.EPOCH_COMPLETED)
            - "iteration", after completion of each iteration (equivalent of ignite.engine.Events.ITERATION_COMPLETED)
            - any ignite built-in event from ignite.engine.Events.
            Defaults to "epoch".
        log_level: log level (integer) for some garbage collection information as below. Defaults to 10 (DEBUG).
            - 50 (CRITICAL)
            - 40 (ERROR)
            - 30 (WARNING)
            - 20 (INFO)
            - 10 (DEBUG)
            - 0 (NOTSET)
    """

    def __init__(self, trigger_event: str | Events | CallableEventWithFilter = "epoch", log_level: int = 10):
        self.trigger_event: Events | CallableEventWithFilter
        if isinstance(trigger_event, (Events, CallableEventWithFilter)):
            self.trigger_event = trigger_event
        elif trigger_event.lower() == "epoch":
            self.trigger_event = Events.EPOCH_COMPLETED
        elif trigger_event.lower() == "iteration":
            self.trigger_event = Events.ITERATION_COMPLETED
        else:
            raise ValueError(
                f"'trigger_event' should be either epoch, iteration, or an ignite built-in event from"
                f" ignite.engine.Events, '{trigger_event}' was given."
            )

        self.log_level = log_level

    def attach(self, engine: Engine) -> None:
        if not engine.has_event_handler(self, self.trigger_event):
            engine.add_event_handler(self.trigger_event, self)

    def __call__(self, engine: Engine) -> None:
        """
        This method calls python garbage collector.

        Args:
            engine: Ignite Engine, it should be either a trainer or validator.
        """
        # get count before garbage collection
        pre_count = gc.get_count()
        # first call to garbage collector
        gc.collect()
        # second call to garbage collector
        unreachable = gc.collect()
        # get count after garbage collection
        after_count = gc.get_count()
        engine.logger.log(
            self.log_level,
            f"Garbage Count: [before: {pre_count}] -> [after: {after_count}] (unreachable : {unreachable})",
        )



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Profiling MetaTensor
"""

from __future__ import annotations

import cProfile

import torch

from monai.data.meta_tensor import MetaTensor

if __name__ == "__main__":
    n_chan = 3
    for hwd in (10, 200):
        shape = (n_chan, hwd, hwd, hwd)
        a = MetaTensor(torch.rand(shape), meta={"affine": torch.eye(4) * 2, "fname": "something1"})
        b = MetaTensor(torch.rand(shape), meta={"affine": torch.eye(4) * 3, "fname": "something2"})
        cProfile.run("c = a + b", filename=f"out_{hwd}.prof")



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
import torch

from monai.handlers import ROCAUC
from monai.transforms import Activations, AsDiscrete


class TestHandlerROCAUC(unittest.TestCase):

    def test_compute(self):
        auc_metric = ROCAUC()
        act = Activations(softmax=True)
        to_onehot = AsDiscrete(to_onehot=2)

        y_pred = [torch.Tensor([0.1, 0.9]), torch.Tensor([0.3, 1.4])]
        y = [torch.Tensor([0]), torch.Tensor([1])]
        y_pred = [act(p) for p in y_pred]
        y = [to_onehot(y_) for y_ in y]
        auc_metric.update([y_pred, y])

        y_pred = [torch.Tensor([0.2, 0.1]), torch.Tensor([0.1, 0.5])]
        y = [torch.Tensor([0]), torch.Tensor([1])]
        y_pred = [act(p) for p in y_pred]
        y = [to_onehot(y_) for y_ in y]

        auc_metric.update([y_pred, y])

        auc = auc_metric.compute()
        np.testing.assert_allclose(0.75, auc)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch
from parameterized import parameterized

from monai.networks import eval_mode
from monai.networks.nets import SegResNetDS, SegResNetDS2
from tests.test_utils import SkipIfBeforePyTorchVersion, test_script_save

device = "cuda" if torch.cuda.is_available() else "cpu"
TEST_CASE_SEGRESNET_DS = []
for spatial_dims in range(2, 4):
    for init_filters in [8, 16]:
        for act in ["relu", "leakyrelu"]:
            for norm in ["BATCH", ("instance", {"affine": True})]:
                for upsample_mode in ["deconv", "nontrainable"]:
                    test_case = [
                        {
                            "spatial_dims": spatial_dims,
                            "init_filters": init_filters,
                            "act": act,
                            "norm": norm,
                            "upsample_mode": upsample_mode,
                        },
                        (2, 1, *([16] * spatial_dims)),
                        (2, 2, *([16] * spatial_dims)),
                    ]
                    TEST_CASE_SEGRESNET_DS.append(test_case)

TEST_CASE_SEGRESNET_DS2 = []
for spatial_dims in range(2, 4):
    for out_channels in [1, 2]:
        for dsdepth in [1, 2, 3]:
            test_case = [
                {"spatial_dims": spatial_dims, "init_filters": 8, "out_channels": out_channels, "dsdepth": dsdepth},
                (2, 1, *([16] * spatial_dims)),
                (2, out_channels, *([16] * spatial_dims)),
            ]
            TEST_CASE_SEGRESNET_DS2.append(test_case)

TEST_CASE_SEGRESNET_DS3 = [
    ({"init_filters": 8, "dsdepth": 2, "resolution": None}, (2, 1, 16, 16, 16), ((2, 2, 16, 16, 16), (2, 2, 8, 8, 8))),
    (
        {"init_filters": 8, "dsdepth": 3, "resolution": None},
        (2, 1, 16, 16, 16),
        ((2, 2, 16, 16, 16), (2, 2, 8, 8, 8), (2, 2, 4, 4, 4)),
    ),
    (
        {"init_filters": 8, "dsdepth": 3, "resolution": [1, 1, 5]},
        (2, 1, 16, 16, 16),
        ((2, 2, 16, 16, 16), (2, 2, 8, 8, 16), (2, 2, 4, 4, 16)),
    ),
    (
        {"init_filters": 8, "dsdepth": 3, "resolution": [1, 2, 5]},
        (2, 1, 16, 16, 16),
        ((2, 2, 16, 16, 16), (2, 2, 8, 8, 16), (2, 2, 4, 8, 16)),
    ),
]


class TestSegResNetDS(unittest.TestCase):
    @parameterized.expand(TEST_CASE_SEGRESNET_DS)
    def test_shape(self, input_param, input_shape, expected_shape):
        net = SegResNetDS(**input_param).to(device)
        with eval_mode(net):
            result = net(torch.randn(input_shape).to(device))
            self.assertEqual(result.shape, expected_shape, msg=str(input_param))

    @parameterized.expand(TEST_CASE_SEGRESNET_DS)
    def test_shape_ds2(self, input_param, input_shape, expected_shape):
        net = SegResNetDS2(**input_param).to(device)
        with eval_mode(net):
            result = net(torch.randn(input_shape).to(device), with_label=False)
            self.assertEqual(result[0].shape, expected_shape, msg=str(input_param))
            self.assertTrue(result[1] == [])

            result = net(torch.randn(input_shape).to(device), with_point=False)
            self.assertEqual(result[1].shape, expected_shape, msg=str(input_param))
            self.assertTrue(result[0] == [])

    @parameterized.expand(TEST_CASE_SEGRESNET_DS2)
    def test_shape2(self, input_param, input_shape, expected_shape):
        dsdepth = input_param.get("dsdepth", 1)
        for net in [SegResNetDS, SegResNetDS2]:
            net = net(**input_param).to(device)
            net.train()
            if isinstance(net, SegResNetDS2):
                result = net(torch.randn(input_shape).to(device), with_label=False)[0]
            else:
                result = net(torch.randn(input_shape).to(device))
            if dsdepth > 1:
                assert isinstance(result, list)
                self.assertEqual(dsdepth, len(result))
                for i in range(dsdepth):
                    self.assertEqual(
                        result[i].shape,
                        expected_shape[:2] + tuple(e // (2**i) for e in expected_shape[2:]),
                        msg=str(input_param),
                    )
            else:
                assert isinstance(result, torch.Tensor)
                self.assertEqual(result.shape, expected_shape, msg=str(input_param))

            if not isinstance(net, SegResNetDS2):
                # eval mode of SegResNetDS2 has same output as training mode
                # so only test eval mode for SegResNetDS
                net.eval()
                result = net(torch.randn(input_shape).to(device))
                assert isinstance(result, torch.Tensor)
                self.assertEqual(result.shape, expected_shape, msg=str(input_param))

    @parameterized.expand(TEST_CASE_SEGRESNET_DS3)
    def test_shape3(self, input_param, input_shape, expected_shapes):
        dsdepth = input_param.get("dsdepth", 1)
        for net in [SegResNetDS, SegResNetDS2]:
            net = net(**input_param).to(device)
            net.train()
            if isinstance(net, SegResNetDS2):
                result = net(torch.randn(input_shape).to(device), with_point=False)[1]
            else:
                result = net(torch.randn(input_shape).to(device))
            assert isinstance(result, list)
            self.assertEqual(dsdepth, len(result))
            for i in range(dsdepth):
                self.assertEqual(result[i].shape, expected_shapes[i], msg=str(input_param))

    def test_ill_arg(self):
        with self.assertRaises(ValueError):
            SegResNetDS(spatial_dims=4)

        with self.assertRaises(ValueError):
            SegResNetDS2(spatial_dims=4)

    @SkipIfBeforePyTorchVersion((1, 10))
    def test_script(self):
        input_param, input_shape, _ = TEST_CASE_SEGRESNET_DS[0]
        net = SegResNetDS(**input_param)
        test_data = torch.randn(input_shape)
        test_script_save(net, test_data)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import torch

from monai.utils.module import optional_import

_C, _ = optional_import("monai._C")

__all__ = ["BilateralFilter", "PHLFilter", "TrainableBilateralFilter", "TrainableJointBilateralFilter"]


class BilateralFilter(torch.autograd.Function):
    """
    Blurs the input tensor spatially whilst preserving edges. Can run on 1D, 2D, or 3D,
    tensors (on top of Batch and Channel dimensions). Two implementations are provided,
    an exact solution and a much faster approximation which uses a permutohedral lattice.

    See:
        https://en.wikipedia.org/wiki/Bilateral_filter
        https://graphics.stanford.edu/papers/permutohedral/

    Args:
        input: input tensor.
        spatial_sigma: the standard deviation of the spatial blur. Higher values can
            hurt performance when not using the approximate method (see fast approx).
        color_sigma: the standard deviation of the color blur. Lower values preserve
            edges better whilst higher values tend to a simple gaussian spatial blur.
        fast approx: This flag chooses between two implementations. The approximate method may
            produce artifacts in some scenarios whereas the exact solution may be intolerably
            slow for high spatial standard deviations.

    Returns:
        output (torch.Tensor): output tensor.
    """

    @staticmethod
    def forward(ctx, input, spatial_sigma=5, color_sigma=0.5, fast_approx=True):
        """autograd forward"""
        ctx.ss = spatial_sigma
        ctx.cs = color_sigma
        ctx.fa = fast_approx
        output_data = _C.bilateral_filter(input, spatial_sigma, color_sigma, fast_approx)
        if torch.cuda.is_available():
            torch.cuda.synchronize()
        return output_data

    @staticmethod
    def backward(ctx, grad_output):
        """autograd backward"""
        spatial_sigma, color_sigma, fast_approx = ctx.ss, ctx.cs, ctx.fa
        grad_input = _C.bilateral_filter(grad_output, spatial_sigma, color_sigma, fast_approx)
        return grad_input, None, None, None


class PHLFilter(torch.autograd.Function):
    """
    Filters input based on arbitrary feature vectors. Uses a permutohedral
    lattice data structure to efficiently approximate n-dimensional gaussian
    filtering. Complexity is broadly independent of kernel size. Most applicable
    to higher filter dimensions and larger kernel sizes.

    See:
        https://graphics.stanford.edu/papers/permutohedral/

    Args:
        input: input tensor to be filtered.
        features: feature tensor used to filter the input.
        sigmas: the standard deviations of each feature in the filter.

    Returns:
        output (torch.Tensor): output tensor.
    """

    @staticmethod
    def forward(ctx, input, features, sigmas=None):
        scaled_features = features
        if sigmas is not None:
            for i in range(features.size(1)):
                scaled_features[:, i, ...] /= sigmas[i]

        ctx.save_for_backward(scaled_features)
        output_data = _C.phl_filter(input, scaled_features)
        return output_data

    @staticmethod
    def backward(ctx, grad_output):
        raise NotImplementedError("PHLFilter does not currently support Backpropagation")
        # scaled_features, = ctx.saved_variables
        # grad_input = _C.phl_filter(grad_output, scaled_features)
        # return grad_input


class TrainableBilateralFilterFunction(torch.autograd.Function):
    """
    torch.autograd.Function for the TrainableBilateralFilter layer.

    See:
        F. Wagner, et al., Ultralow-parameter denoising: Trainable bilateral filter layers in
        computed tomography, Medical Physics (2022), https://doi.org/10.1002/mp.15718

    Args:
        input: input tensor to be filtered.
        sigma x: trainable standard deviation of the spatial filter kernel in x direction.
        sigma y: trainable standard deviation of the spatial filter kernel in y direction.
        sigma z: trainable standard deviation of the spatial filter kernel in z direction.
        color sigma: trainable standard deviation of the intensity range kernel. This filter
            parameter determines the degree of edge preservation.

    Returns:
        output (torch.Tensor): filtered tensor.
    """

    @staticmethod
    def forward(ctx, input_img, sigma_x, sigma_y, sigma_z, color_sigma):
        output_tensor, output_weights_tensor, do_dx_ki, do_dsig_r, do_dsig_x, do_dsig_y, do_dsig_z = _C.tbf_forward(
            input_img, sigma_x, sigma_y, sigma_z, color_sigma
        )

        ctx.save_for_backward(
            input_img,
            sigma_x,
            sigma_y,
            sigma_z,
            color_sigma,
            output_tensor,
            output_weights_tensor,
            do_dx_ki,
            do_dsig_r,
            do_dsig_x,
            do_dsig_y,
            do_dsig_z,
        )
        if torch.cuda.is_available():
            torch.cuda.synchronize()
        return output_tensor

    @staticmethod
    def backward(ctx, grad_output):
        input_img = ctx.saved_tensors[0]  # input image
        sigma_x = ctx.saved_tensors[1]
        sigma_y = ctx.saved_tensors[2]
        sigma_z = ctx.saved_tensors[3]
        color_sigma = ctx.saved_tensors[4]
        output_tensor = ctx.saved_tensors[5]  # filtered image
        output_weights_tensor = ctx.saved_tensors[6]  # weights
        do_dx_ki = ctx.saved_tensors[7]  # derivative of output with respect to input, while k==i
        do_dsig_r = ctx.saved_tensors[8]  # derivative of output with respect to range sigma
        do_dsig_x = ctx.saved_tensors[9]  # derivative of output with respect to sigma x
        do_dsig_y = ctx.saved_tensors[10]  # derivative of output with respect to sigma y
        do_dsig_z = ctx.saved_tensors[11]  # derivative of output with respect to sigma z

        # calculate gradient with respect to the sigmas
        grad_color_sigma = torch.sum(grad_output * do_dsig_r)
        grad_sig_x = torch.sum(grad_output * do_dsig_x)
        grad_sig_y = torch.sum(grad_output * do_dsig_y)
        grad_sig_z = torch.sum(grad_output * do_dsig_z)

        grad_output_tensor = _C.tbf_backward(
            grad_output,
            input_img,
            output_tensor,
            output_weights_tensor,
            do_dx_ki,
            sigma_x,
            sigma_y,
            sigma_z,
            color_sigma,
        )

        return grad_output_tensor, grad_sig_x, grad_sig_y, grad_sig_z, grad_color_sigma


class TrainableBilateralFilter(torch.nn.Module):
    """
    Implementation of a trainable bilateral filter layer as proposed in the corresponding publication.
    All filter parameters can be trained data-driven. The spatial filter kernels x, y, and z determine
    image smoothing whereas the color parameter specifies the amount of edge preservation.
    Can run on 1D, 2D, or 3D tensors (on top of Batch and Channel dimensions).

    See:
        F. Wagner, et al., Ultralow-parameter denoising: Trainable bilateral filter layers in
        computed tomography, Medical Physics (2022), https://doi.org/10.1002/mp.15718

    Args:
        input: input tensor to be filtered.
        spatial_sigma: tuple (sigma_x, sigma_y, sigma_z) initializing the trainable standard
            deviations of the spatial filter kernels. Tuple length must equal the number of
            spatial input dimensions.
        color_sigma: trainable standard deviation of the intensity range kernel. This filter
            parameter determines the degree of edge preservation.

    Returns:
        output (torch.Tensor): filtered tensor.
    """

    def __init__(self, spatial_sigma, color_sigma):
        super().__init__()

        if isinstance(spatial_sigma, float):
            spatial_sigma = [spatial_sigma, spatial_sigma, spatial_sigma]
            self.len_spatial_sigma = 3
        elif len(spatial_sigma) == 1:
            spatial_sigma = [spatial_sigma[0], 0.01, 0.01]
            self.len_spatial_sigma = 1
        elif len(spatial_sigma) == 2:
            spatial_sigma = [spatial_sigma[0], spatial_sigma[1], 0.01]
            self.len_spatial_sigma = 2
        elif len(spatial_sigma) == 3:
            spatial_sigma = [spatial_sigma[0], spatial_sigma[1], spatial_sigma[2]]
            self.len_spatial_sigma = 3
        else:
            raise ValueError(
                f"len(spatial_sigma) {spatial_sigma} must match number of spatial dims {self.ken_spatial_sigma}."
            )

        # Register sigmas as trainable parameters.
        self.sigma_x = torch.nn.Parameter(torch.tensor(spatial_sigma[0]))
        self.sigma_y = torch.nn.Parameter(torch.tensor(spatial_sigma[1]))
        self.sigma_z = torch.nn.Parameter(torch.tensor(spatial_sigma[2]))
        self.sigma_color = torch.nn.Parameter(torch.tensor(color_sigma))

    def forward(self, input_tensor):
        if input_tensor.shape[1] != 1:
            raise ValueError(
                f"Currently channel dimensions >1 ({input_tensor.shape[1]}) are not supported. "
                "Please use multiple parallel filter layers if you want "
                "to filter multiple channels."
            )

        len_input = len(input_tensor.shape)

        # C++ extension so far only supports 5-dim inputs.
        if len_input == 3:
            input_tensor = input_tensor.unsqueeze(3).unsqueeze(4)
        elif len_input == 4:
            input_tensor = input_tensor.unsqueeze(4)

        if self.len_spatial_sigma != len_input:
            raise ValueError(f"Spatial dimension ({len_input}) must match initialized len(spatial_sigma).")

        prediction = TrainableBilateralFilterFunction.apply(
            input_tensor, self.sigma_x, self.sigma_y, self.sigma_z, self.sigma_color
        )

        # Make sure to return tensor of the same shape as the input.
        if len_input == 3:
            prediction = prediction.squeeze(4).squeeze(3)
        elif len_input == 4:
            prediction = prediction.squeeze(4)

        return prediction


class TrainableJointBilateralFilterFunction(torch.autograd.Function):
    """
    torch.autograd.Function for the TrainableJointBilateralFilter layer.

    See:
        F. Wagner, et al., Trainable joint bilateral filters for enhanced prediction stability in
        low-dose CT, Scientific Reports (2022), https://doi.org/10.1038/s41598-022-22530-4

    Args:
        input: input tensor to be filtered.
        guide: guidance image tensor to be used during filtering.
        sigma x: trainable standard deviation of the spatial filter kernel in x direction.
        sigma y: trainable standard deviation of the spatial filter kernel in y direction.
        sigma z: trainable standard deviation of the spatial filter kernel in z direction.
        color sigma: trainable standard deviation of the intensity range kernel. This filter
            parameter determines the degree of edge preservation.

    Returns:
        output (torch.Tensor): filtered tensor.
    """

    @staticmethod
    def forward(ctx, input_img, guidance_img, sigma_x, sigma_y, sigma_z, color_sigma):
        output_tensor, output_weights_tensor, do_dx_ki, do_dsig_r, do_dsig_x, do_dsig_y, do_dsig_z = _C.tjbf_forward(
            input_img, guidance_img, sigma_x, sigma_y, sigma_z, color_sigma
        )

        ctx.save_for_backward(
            input_img,
            sigma_x,
            sigma_y,
            sigma_z,
            color_sigma,
            output_tensor,
            output_weights_tensor,
            do_dx_ki,
            do_dsig_r,
            do_dsig_x,
            do_dsig_y,
            do_dsig_z,
            guidance_img,
        )
        if torch.cuda.is_available():
            torch.cuda.synchronize()
        return output_tensor

    @staticmethod
    def backward(ctx, grad_output):
        input_img = ctx.saved_tensors[0]  # input image
        sigma_x = ctx.saved_tensors[1]
        sigma_y = ctx.saved_tensors[2]
        sigma_z = ctx.saved_tensors[3]
        color_sigma = ctx.saved_tensors[4]
        output_tensor = ctx.saved_tensors[5]  # filtered image
        output_weights_tensor = ctx.saved_tensors[6]  # weights
        do_dx_ki = ctx.saved_tensors[7]  # derivative of output with respect to input, while k==i
        do_dsig_r = ctx.saved_tensors[8]  # derivative of output with respect to range sigma
        do_dsig_x = ctx.saved_tensors[9]  # derivative of output with respect to sigma x
        do_dsig_y = ctx.saved_tensors[10]  # derivative of output with respect to sigma y
        do_dsig_z = ctx.saved_tensors[11]  # derivative of output with respect to sigma z
        guidance_img = ctx.saved_tensors[12]  # guidance image

        # calculate gradient with respect to the sigmas
        grad_color_sigma = torch.sum(grad_output * do_dsig_r)
        grad_sig_x = torch.sum(grad_output * do_dsig_x)
        grad_sig_y = torch.sum(grad_output * do_dsig_y)
        grad_sig_z = torch.sum(grad_output * do_dsig_z)

        grad_output_tensor, grad_guidance_tensor = _C.tjbf_backward(
            grad_output,
            input_img,
            guidance_img,
            output_tensor,
            output_weights_tensor,
            do_dx_ki,
            sigma_x,
            sigma_y,
            sigma_z,
            color_sigma,
        )

        return grad_output_tensor, grad_guidance_tensor, grad_sig_x, grad_sig_y, grad_sig_z, grad_color_sigma


class TrainableJointBilateralFilter(torch.nn.Module):
    """
    Implementation of a trainable joint bilateral filter layer as proposed in the corresponding publication.
    The guidance image is used as additional (edge) information during filtering. All filter parameters and the
    guidance image can be trained data-driven. The spatial filter kernels x, y, and z determine
    image smoothing whereas the color parameter specifies the amount of edge preservation.
    Can run on 1D, 2D, or 3D tensors (on top of Batch and Channel dimensions). Input tensor shape must match
    guidance tensor shape.

    See:
        F. Wagner, et al., Trainable joint bilateral filters for enhanced prediction stability in
        low-dose CT, Scientific Reports (2022), https://doi.org/10.1038/s41598-022-22530-4

    Args:
        input: input tensor to be filtered.
        guide: guidance image tensor to be used during filtering.
        spatial_sigma: tuple (sigma_x, sigma_y, sigma_z) initializing the trainable standard
            deviations of the spatial filter kernels. Tuple length must equal the number of
            spatial input dimensions.
        color_sigma: trainable standard deviation of the intensity range kernel. This filter
            parameter determines the degree of edge preservation.

    Returns:
        output (torch.Tensor): filtered tensor.
    """

    def __init__(self, spatial_sigma, color_sigma):
        super().__init__()

        if isinstance(spatial_sigma, float):
            spatial_sigma = [spatial_sigma, spatial_sigma, spatial_sigma]
            self.len_spatial_sigma = 3
        elif len(spatial_sigma) == 1:
            spatial_sigma = [spatial_sigma[0], 0.01, 0.01]
            self.len_spatial_sigma = 1
        elif len(spatial_sigma) == 2:
            spatial_sigma = [spatial_sigma[0], spatial_sigma[1], 0.01]
            self.len_spatial_sigma = 2
        elif len(spatial_sigma) == 3:
            spatial_sigma = [spatial_sigma[0], spatial_sigma[1], spatial_sigma[2]]
            self.len_spatial_sigma = 3
        else:
            raise ValueError(
                f"len(spatial_sigma) {spatial_sigma} must match number of spatial dims {self.ken_spatial_sigma}."
            )

        # Register sigmas as trainable parameters.
        self.sigma_x = torch.nn.Parameter(torch.tensor(spatial_sigma[0]))
        self.sigma_y = torch.nn.Parameter(torch.tensor(spatial_sigma[1]))
        self.sigma_z = torch.nn.Parameter(torch.tensor(spatial_sigma[2]))
        self.sigma_color = torch.nn.Parameter(torch.tensor(color_sigma))

    def forward(self, input_tensor, guidance_tensor):
        if input_tensor.shape[1] != 1:
            raise ValueError(
                f"Currently channel dimensions >1 ({input_tensor.shape[1]}) are not supported. "
                "Please use multiple parallel filter layers if you want "
                "to filter multiple channels."
            )
        if input_tensor.shape != guidance_tensor.shape:
            raise ValueError(
                "Shape of input image must equal shape of guidance image."
                f"Got {input_tensor.shape} and {guidance_tensor.shape}."
            )

        len_input = len(input_tensor.shape)

        # C++ extension so far only supports 5-dim inputs.
        if len_input == 3:
            input_tensor = input_tensor.unsqueeze(3).unsqueeze(4)
            guidance_tensor = guidance_tensor.unsqueeze(3).unsqueeze(4)
        elif len_input == 4:
            input_tensor = input_tensor.unsqueeze(4)
            guidance_tensor = guidance_tensor.unsqueeze(4)

        if self.len_spatial_sigma != len_input:
            raise ValueError(f"Spatial dimension ({len_input}) must match initialized len(spatial_sigma).")

        prediction = TrainableJointBilateralFilterFunction.apply(
            input_tensor, guidance_tensor, self.sigma_x, self.sigma_y, self.sigma_z, self.sigma_color
        )

        # Make sure to return tensor of the same shape as the input.
        if len_input == 3:
            prediction = prediction.squeeze(4).squeeze(3)
        elif len_input == 4:
            prediction = prediction.squeeze(4)

        return prediction
